#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/rman.h>
#include <sys/conf.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/mman.h>
#include <sys/proc.h>

#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>

#include <dev/iicbus/iiconf.h>
#include "iicbus_if.h"

#include <vm/vm.h>
#include <vm/vm_extern.h>
#include <vm/pmap.h>                /* req. by machine/pmap.h       */
#include <machine/pmap.h>           /* vtophys                      */
#include <vm/vm_map.h>
#include <vm/vm_object.h>
#include <vm/vm_kern.h>
#include "saa713x_reg.h"
#include "saa713x_ioctl.h"
#include "saa713x_var.h"
#include "saa713x_audio.h"
#include "saa713x_ioctl.h"

MALLOC_DECLARE(M_SAA_ACAP);
MALLOC_DEFINE(M_SAA_ACAP, "adocapbuf", "SAA713x Audio capture buffer");

/* 
 * About VLock:
 *	VLock garranties to generate only 
 *	ACPF samples per feild, hence no
 *	drift.
 */

static d_open_t		saa_audiodev_open;
static d_close_t	saa_audiodev_close;
static d_ioctl_t	saa_audiodev_ioctl;
static d_mmap_t		saa_audiodev_mmap;

static struct cdevsw saa_audio_cdevsw = {
	.d_name 	= "sau",
	.d_open 	= saa_audiodev_open,
	.d_close 	= saa_audiodev_close,
	.d_ioctl 	= saa_audiodev_ioctl,
	.d_mmap		= saa_audiodev_mmap,
	DVERSION
	DFLAGS
};

static struct saa_audio_opt saa_default_audio_opt = { /* PAL A2 FM Defaults */
	.sao_flag_idarea = 0,
	.sao_idmod = 0,

	.sao_ch1mode = CH1MODE_FM,
	.sao_ch2mode = CH2MODE_FM,
	.sao_fltbw = FLTBW_C1MIN_C2MIN,

	.sao_siffreq1 = 0x729555,
	.sao_siffreq2 = 0x77A101,

	.sao_flag_hpen = 0,
	.sao_sifp = 0,
	.sao_sifa = SIFA_N6DB,

	.sao_sifgain = 0,
	.sao_flag_sifagc_off = 0,
	.sao_flag_sifagc_slow = 0,

	.sao_flag_fmdsch1 = 0,
	.sao_fmdsch1 = 0,
	.sao_flag_fmdsch2 = 0,
	.sao_fmdsch2 = 0,

	.sao_flag_idswfm = 1,		/* Auto DMSel ON */
	.sao_fmdmsel = 0,

	.sao_dac_sdos = SDOS_FMAM_FMAM,
	.sao_dac_avl = AVL_OFF,
	.sao_dac_ogain = SDGS_9DB,

	.sao_flag_aasdma = 1,
	.sao_dsp_csm = CSM_LEFT_RIGHT,

	.sao_flag_ch1_2vrms = 0,
	.sao_flag_ch2_2vrms = 0
};


static void
audio_dma_cb(void *arg, bus_dma_segment_t *segs, int nsegs, int error)
{
	struct saa_softc *sc = (struct saa_softc*)arg;
	int i;
	
	if (error == EFBIG) {
		sc->sc_capbuf_loaderror = EFBIG;
		wakeup(sc);
		return;
	}	

	/* Invariant loadidx < 1024 */
	i = 0;
	while (i < nsegs) {
	   sc->sc_capbuf_pgsegs[2][sc->sc_capbuf_loadidx++] =
				segs[i];
	   sc->sc_capbuf_loadlen += segs[i++].ds_len;
	}
	sc->sc_audio_capbuf_loadcount += nsegs;

	if (sc->sc_capbuf_loadlen >= sc->sc_audio_capbuf_sz) 
		wakeup(sc);
}


static int
saa_audio_capbuf_init(struct saa_softc *sc)
{
	int rval;

	if (sc->sc_audio_capbuf == NULL) {
		sc->sc_audio_capbuf_sz = SAA_ACAP_FRAME_MAX_SZ ;
		sc->sc_audio_capbuf = malloc(sc->sc_audio_capbuf_sz,
						 M_SAA_ACAP, M_WAITOK | M_ZERO);
		if (sc->sc_audio_capbuf == NULL) {
			device_printf(sc->sc_dev, "Error malloc'ing audio "
						"capture buffer nr");
			return ENOMEM;
		}
		if (bus_dma_tag_create(0, 
				4096, 0,
				BUS_SPACE_MAXADDR_32BIT,
				BUS_SPACE_MAXADDR,
				0, 0,
				sc->sc_audio_capbuf_sz, 
				1024,
				sc->sc_audio_capbuf_sz,
				0,	
				0, 0,
				&sc->sc_audio_capbuf_tag) != 0) {
			device_printf(sc->sc_dev, "Error creating audio "
						"capture buffer\n");
			return EAGAIN;
		}
		if (bus_dmamap_create(sc->sc_audio_capbuf_tag,
					0,
					&sc->sc_audio_capbuf_map) != 0) {
			device_printf(sc->sc_dev, "Error creating audio "
					"capture buffer map\n");
			return EAGAIN;
		}
		sc->sc_audio_capbuf_map_destroyflag = 1;
		sc->sc_audio_capbuf_loadcount = 0;
		sc->sc_capbuf_loadlen = 0;
		sc->sc_capbuf_loadidx = 0;
		sc->sc_capbuf_loaderror = 0;
		rval = bus_dmamap_load(sc->sc_audio_capbuf_tag,
					sc->sc_audio_capbuf_map,
					sc->sc_audio_capbuf, 
					sc->sc_audio_capbuf_sz,
					audio_dma_cb, sc,
					0);
		if (rval == EINPROGRESS) {
			tsleep(sc, PRI_MAX_KERN, "dmapld", 0);
		} else if (rval != 0) {
			device_printf(sc->sc_dev, "Error loading audio "
					 	"capture buffer\n");
			return rval;
		}

		if (sc->sc_capbuf_loaderror != 0) {
			device_printf(sc->sc_dev, "Error, callback failed "
					"to load audio capture buffer "
					"(idx: %d)\n", 
					sc->sc_audio_capbuf_loadcount);
			return EAGAIN;
		}
		sc->sc_audio_capbuf_map_unloadflag = 1;
	}

	saa_build_pta(sc, 
			sc->sc_capbuf_pgsegs[2], 
			sc->sc_audio_capbuf_loadcount,
			&(sc->sc_pta[SAA_CHAN_AUDIO * 1024]));

	return 0;
}


static int
saa_audio_capbuf_dnit(struct saa_softc *sc)
{
	int rval = 0;

	/* flags required as capbuf_map can legally be equal
	 * to zero
	 */
	if (sc->sc_audio_capbuf_map_unloadflag) 
		bus_dmamap_unload(sc->sc_audio_capbuf_tag,
					sc->sc_audio_capbuf_map);
	
	if (sc->sc_audio_capbuf_map_destroyflag)
		rval = bus_dmamap_destroy(
					sc->sc_audio_capbuf_tag,
					sc->sc_audio_capbuf_map);	

	if (!rval && sc->sc_audio_capbuf) {
		free(sc->sc_audio_capbuf, M_SAA_ACAP);
		sc->sc_audio_capbuf = 0;
	}

	if (!rval && sc->sc_audio_capbuf_tag)
	    rval = bus_dma_tag_destroy(sc->sc_audio_capbuf_tag);

	return rval;
}


static
void saa_audio_setopt(struct saa_softc *sc, struct saa_audio_opt *o)
{
	REG_WRITEB(BREG_DXCO_CTRL, 
			(o->sao_flag_idarea ? DXCO_CTRL_IDAREA : 0) |
			o->sao_idmod);

	REG_WRITEB(BREG_DMOD_CFG,
		  	o->sao_ch1mode | o->sao_fltbw |
			(o->sao_ch2mode & 0x1) | 
			((o->sao_ch2mode & 0x2) << 
					(DMOD_CFG_CH2_MODE_1_SHIFT - 1)));

	REG_WRITE(REG_CAR1FREQ, o->sao_siffreq1);
	REG_WRITE(REG_CAR2FREQ, o->sao_siffreq2);

	REG_WRITEB(BREG_SFREQ,
		(o->sao_flag_hpen ? SFREQ_HPEN : 0) | 
		((o->sao_sifp << SFREQ_SIFP_SHIFT) & SFREQ_SIFP_MASK) |
		o->sao_sifa);
	
	REG_WRITEB(BREG_AUDGAIN,
		((o->sao_sifgain << AUDGAIN_SIFG_SHIFT) & AUDGAIN_SIFG_MASK) |
		(o->sao_flag_sifagc_off ? AUDGAIN_AGC_OFF : 0) |
		(o->sao_flag_sifagc_slow ? AUDGAIN_AGC_SLOW : 0));

	REG_WRITEB(BREG_FMDEMP,
		(o->sao_flag_fmdsch1 ? FMDEMP_ADEMP1 : 0) |
		o->sao_fmdsch1 |
		(o->sao_flag_fmdsch2 ? FMDEMP_ADEMP2 : 0) |
		o->sao_fmdsch2);

	REG_WRITEB(BREG_FMDMAT,
		(o->sao_flag_idswfm ? FMDMAT_IDSWFM : 0) |
		o->sao_fmdmsel);

	REG_WRITEB(BREG_DACOSEL,
		o->sao_dac_sdos | o->sao_dac_avl | o->sao_dac_ogain);

	REG_WRITEB(BREG_DSPOSEL, 
		(o->sao_flag_aasdma ? DSPOSEL_AASDMA : 0) |
		o->sao_dsp_csm);

	if (o->sao_flag_ch1_2vrms)
		REG_SETB(BREG_AUDIOSEL, AUDIOSEL_VSEL1_2VRMS);
	else
		REG_CLRB(BREG_AUDIOSEL, AUDIOSEL_VSEL1_2VRMS);

	if (o->sao_flag_ch2_2vrms)
		REG_SETB(BREG_AUDIOSEL, AUDIOSEL_VSEL2_2VRMS);
	else
		REG_CLRB(BREG_AUDIOSEL, AUDIOSEL_VSEL2_2VRMS);
}

 
int
saa_audio_init(struct saa_softc *sc)
{
	int unit, rval;

	unit = device_get_unit(sc->sc_dev);
	sc->sc_audio_opt = saa_default_audio_opt;
	saa_audio_setopt(sc, &saa_default_audio_opt);
	if ((rval = saa_audio_capbuf_init(sc)) != 0)
		return rval;
	sc->sc_audio_capbuf_allocflag = 1;
	sc->sc_audio_cdev = make_dev(&saa_audio_cdevsw, 
					unit,
					UID_ROOT, GID_WHEEL, 0600,
					"sau%d", unit);
	REG_SET(REG_MCR, (MCR_EXOSC | MCR_APLLE | MCR_ESFE | 
				MCR_EBDAC | MCR_EBADC));
	return 0;
}


int
saa_audio_dnit(struct saa_softc *sc)
{

	if (sc->sc_audio_capbuf_allocflag) {
		if (saa_audio_capbuf_dnit(sc) != 0)
			device_printf(sc->sc_dev,
					"device_close: error releasing"
					" audio capture buffer. Mem. leak\n");
		else
			sc->sc_audio_capbuf_allocflag = 0;	

	}
	if (sc->sc_audio_cdev)
		destroy_dev(sc->sc_audio_cdev);

	return 0;
}


static int
saa_audio_capture_off(struct saa_softc *sc)
{
	REG_CLR(REG_FIEN, (FIEN_SLICE3_0 | FIEN_SLICE3_1));
	REG_CLR(REG_MCR, MCR_CHAN6_TE);
	REG_WRITEB(BREG_I2SOFMT, 0);
	sc->sc_audio_flags &= ~SCAF_ON;
	return 0;
}


static int
saa_audio_capture_on(struct saa_softc *sc, struct saa_audio_spec *s)
{
	int ctrl;

	/* XXX check sampcount * 2, as we have two buffers minimum */
	if (s->sas_cap_sampcount == 0 || 
		(s->sas_cap_sampcount * 2) > SAA_ACAP_FRAME_MAX_SZ)
		return ENOSPC;

	switch (s->sas_sfs) {
	case SFS_SIF_32KHZ:
	case SFS_BBAND_32KHZ:
	case SFS_BBAND_48KHZ:
		if (s->sas_aclkfreq == ACLKFREQ_32p11MHZ) {
			REG_WRITEW(WREG_ACNI_0_15, 0x187DE7 & 0xffff);
			REG_WRITEB(BREG_ACNI_16_21, (0x187DE7 >> 16) & 0x3f);
		} else {
			REG_WRITEW(WREG_ACNI_0_15, 0x200000 & 0xffff);
			REG_WRITEB(BREG_ACNI_16_21, (0x200000 >> 16) & 0x3f);
		}
		if (s->sas_flags & ASPEC_FLAG_NTSC) {
			REG_WRITEW(WREG_ACPF_0_15, 0x19066 & 0xffff);
			REG_WRITEB(BREG_ACPF_16_17, (0x19066 >> 16) & 0x3);
		} else {
			REG_WRITEW(WREG_ACPF_0_15, 0x1E000 & 0xffff);
			REG_WRITEB(BREG_ACPF_16_17, (0x1E000 >> 16) & 0x3);

		}
		break;
	case SFS_BBAND_44KHZ:
		if (s->sas_aclkfreq == ACLKFREQ_32p11MHZ) {
			REG_WRITEW(WREG_ACNI_0_15, 0x1C2097 & 0xffff);
			REG_WRITEB(BREG_ACNI_16_21, (0x1C2097 >> 16) & 0x3f);
		} else {
			REG_WRITEW(WREG_ACNI_0_15, 0x24C000 & 0xffff);
			REG_WRITEB(BREG_ACNI_16_21, (0x24C000 >> 16) & 0x3f);
		}
		if (s->sas_flags & ASPEC_FLAG_NTSC) {
			REG_WRITEW(WREG_ACPF_0_15, 0x1CBD5 & 0xffff);
			REG_WRITEB(BREG_ACPF_16_17, (0x1CBD5 >> 16) & 0x3);
		} else {
			REG_WRITEW(WREG_ACPF_0_15, 0x22740 & 0xffff);
			REG_WRITEB(BREG_ACPF_16_17, (0x22740 >> 16) & 0x3);
		}
		break;
	default:
		return EINVAL;
	}

	if (s->sas_sfs == SFS_SIF_32KHZ)
		REG_WRITEB(BREG_APLLCTRL, APLLCTRL_SWLOOP);
	else {
		REG_WRITEB(BREG_I2SOFMT, I2SOFMT_SOUT);
		REG_WRITEB(BREG_APLLCTRL, 
			APLLCTRL_SWLOOP |
			((s->sas_flags & ASPEC_FLAG_VLOCK) 
							? 0 : APLLCTRL_APLL));
	}
	
	/* Check and set NOS */
	if (s->sas_icsch)
		REG_SETB(BREG_AUDIOSEL, AUDIOSEL_ICS);
	else
		REG_CLRB(BREG_AUDIOSEL, AUDIOSEL_ICS);

	REG_SETMB(BREG_SFREQ, SFREQ_SFS_MASK, s->sas_sfs);

	REG_WRITEB(BREG_AUDFMT, 
		(s->sas_audfmt | 
		(s->sas_flag_samp2complement ? AUDFMT_FLAG_2COMPLEMENT : 0) |
		s->sas_audsamp |
		s->sas_audsel));

	REG_WRITEB(BREG_AUDMUTE, 0);
	REG_WRITEW(WREG_AUDDMACNT_0_15, ((s->sas_cap_sampcount - 1) & 0xffff));
	REG_WRITEB(BREG_AUDDMACNT_16_21, ((s->sas_cap_sampcount - 1) >> 16) 
									& 0x3f);

	REG_WRITE(REG_DMA_BASE0(SAA_CHAN_AUDIO), 0);
	REG_WRITE(REG_DMA_BASE1(SAA_CHAN_AUDIO), s->sas_cap_sampcount);

	REG_WRITE(REG_DMA_PITCH(SAA_CHAN_AUDIO), 0);
	ctrl = 	(0x1 << DMA_CTRL_BURST_SHIFT) & DMA_CTRL_BURST_MASK;
	ctrl |= DMA_CTRL_ME;
	ctrl |= ((sc->sc_pta_paddr[SAA_CHAN_AUDIO] >> 12) & DMA_CTRL_PTA_MASK);
	REG_WRITE(REG_DMA_CTRL(SAA_CHAN_AUDIO), ctrl);

	sc->sc_sts->ssts_audio_framesz = s->sas_cap_sampcount;
	sc->sc_sts->ssts_audio_sig_count = 0;
	sc->sc_sts->ssts_audio_framectr = 0;
	/* 
	 * mb_cur has to be 1 as on first interrupt we want to flip BASE0
	 * in saa_audio_multibuf_switchbuf
	 */
	sc->sc_sts->ssts_audio_mb_cur = 1;
	sc->sc_sts->ssts_audio_mb_count = 1;

	if (s->sas_flags & ASPEC_FLAG_CAPTURE_MULTIBUF) {
		sc->sc_sts->ssts_audio_mb_count = SAA_ACAP_FRAME_MAX_SZ / 
						sc->sc_sts->ssts_audio_framesz;
		if ((sc->sc_sts->ssts_audio_mb_count > 
						s->sas_cap_mb_limit) &&
		    (s->sas_cap_mb_limit > 0))
		    sc->sc_sts->ssts_audio_mb_count = s->sas_cap_mb_limit;
	}

	REG_SET(REG_MCR, MCR_CHAN6_TE);
#if NOT_YET
	/* 
	 * Interrupts are always enabled, ie.,
	 * even when ASPEC_FLAG_CAPTURE_SIGNAL is not specified as
	 * we always need to update status (frame_ctr, multibuf, etc.)
	 * when capture is on.
	 * If ASPEC_FLAG_CAPTURE_SIGNAL is not set, saa_intr will not 
	 * deliver video signals.
	 *
	 * Feb/02/2005 Not implemented as it will increase the nr. 
	 * of interrupts generated by the device.
	 */
#endif
	if (s->sas_flags & ASPEC_FLAG_CAPTURE_SIGNAL) {
		REG_SET(REG_FIEN,  FIEN_SLICE3_0 | FIEN_SLICE3_1);
	}
	sc->sc_audio_flags |= SCAF_ON;
	return 0;
}


static int 
saa_audiodev_open(DPTYPE dev, int oflags, int dtype, struct thread *td)
{
	struct saa_softc *sc;
	int unit;

	unit = minor(dev);
	sc = devclass_get_softc(saa_devclass, unit);
	//device_busy(devclass_get_device(saa_devclass, unit));
	return 0;
}


static int 
saa_audiodev_close(DPTYPE dev, int fflags, int dtype, struct thread *td)
{
	struct saa_softc *sc;
	int unit;

	unit = minor(dev);
	sc = devclass_get_softc(saa_devclass, unit);
	if (sc->sc_audio_flags & SCAF_ON)
		saa_audio_capture_off(sc);
	//device_unbusy(devclass_get_device(saa_devclass, unit));
	return 0;
}


static int 
saa_audiodev_ioctl(DPTYPE dev, u_long cmd,  caddr_t data, int fflags, 
		struct thread *td)
{
	struct saa_softc *sc;
	int unit, rval = 0;

	unit = minor(dev);
	sc = devclass_get_softc(saa_devclass, unit);

	switch(cmd) {
	case ACAPTUREON:
	{
		struct saa_audio_spec *s = (struct saa_audio_spec*) data;
		if (sc->sc_audio_flags & SCAF_ON) {
			rval = EBUSY;
			break;
		}
		sc->sc_audio_spec = *s;
		if (s->sas_flags & ASPEC_FLAG_CAPTURE_SIGNAL)
			sc->sc_audio_capbuf_sigpid = td->td_proc->p_pid;
		else
			sc->sc_audio_capbuf_sigpid = 0;
		rval =  saa_audio_capture_on(sc, s);
		break;
	}
	case ACAPTUREOFF:
	{
		if (!(sc->sc_audio_flags & SCAF_ON))
			break;
		rval = saa_audio_capture_off(sc);
		sc->sc_audio_capbuf_sigpid = 0;
		break;
	}
	case GETAUDIOOPT:
	{
		struct saa_audio_opt *o = (struct saa_audio_opt*) data;
		*o = sc->sc_audio_opt;
		break;
	}
	case SETAUDIOOPT:
	{
		struct saa_audio_opt *o = (struct saa_audio_opt*) data;
		sc->sc_audio_opt = *o;
		saa_audio_setopt(sc, o);
		break;
	}
	case GETSIFAUDIOLVL:
		*data = REG_READB(BREG_IDENT) & IDENT_SIFL_MASK;
		break;
	case GETFMAUDIOSTS:
		*data = REG_READB(BREG_IDENT) & (~IDENT_SIFL_MASK);
		break;
	case GETAUDIOCH1LVL:
		*data = REG_READB(BREG_CH1_LVLADJ);
		break;
	case SETAUDIOCH1LVL:
		REG_WRITEB(BREG_CH1_LVLADJ, *data);
		break;
	case GETAUDIOCH2LVL:
		*data = REG_READB(BREG_CH2_LVLADJ);
		break;
	case SETAUDIOCH2LVL:
		REG_WRITEB(BREG_CH2_LVLADJ, *data);
		break;
	case GETAUDIOMIXER:
		(*(enum OCS*)data) = 
			REG_READB(BREG_AUDIOSEL) & AUDIOSEL_OCS_MASK;
		break;
	case SETAUDIOMIXER:
		REG_SETMB(BREG_AUDIOSEL, AUDIOSEL_OCS_MASK, *data);
		break;
	case SETAUDIOMUTE:
		if (*data)
			REG_SETB(BREG_AUDMUTE, AUDMUTE_SOUT);
		else
			REG_CLRB(BREG_AUDMUTE, AUDMUTE_SOUT);
		break;
	case SETMONITORCTRL:
	{
		struct saa_audio_monctrl *m = (struct saa_audio_monctrl*)data;
		REG_WRITEB(BREG_MONSEL,
			m->sam_mss | m->sam_mcsm | 
			(m->sam_flag_peak ? MONSEL_PEAK : 0));
		break;
	}
	case GETMONITORLVL:	
		*((int*)data) = REG_READW(WREG_MON_RDOUT);
		break;
	case GETAUDIOVPTR:
	{
		uint32_t *o = (uint32_t*)data;
		*o = REG_READ(REG_SLICE3_STS);
		break;
	}
	case GETGPIODIR:
		*(uint32_t*)data = REG_READ(REG_GPIO_DIR);
		break;
	case SETGPIODIR:
		REG_WRITE(REG_GPIO_DIR, *(uint32_t*)data);
		break;
	case GETGPIODATA:
		*(uint32_t*)data = REG_READ(REG_GPIO_DATA);
		break;
	case SETGPIODATA:
		REG_WRITE(REG_GPIO_DATA, *(uint32_t*)data);
		break;
	default:
		rval = EINVAL;
	}
	return rval;
}


static int 
saa_audiodev_mmap(DPTYPE dev, vm_offset_t offset, vm_paddr_t *paddr, int nprot)
{
	struct saa_softc *sc;
	int unit;
	uint32_t *pta;

	unit = minor(dev);
	sc = devclass_get_softc(saa_devclass, unit);

	if (offset == SAA_MMAP_STS_OFFSET) {
		/* Special case, status page map */
		/* sts is malloced on a page boundary (in video_init) */

		*paddr = vtophys(sc->sc_sts);
		return 0;
	}

	if (nprot & (PROT_EXEC | PROT_WRITE))
		return EPERM;

	if ((offset >= SAA_MMAP_ACAP_OFFSET) &&
		   (offset < (SAA_MMAP_ACAP_OFFSET + SAA_MMAP_ACAP_SZ))) {
		offset -= SAA_MMAP_ACAP_OFFSET;
		offset >>= 12;
	} else {
		return EINVAL;
	}

	if (sc->sc_audio_capbuf_allocflag == 0) {
		return EDOOFUS;
	}
	if (offset > (SAA_MMAP_ACAP_SZ - 4096)) {
		device_printf(sc->sc_dev, "Invalid mmap offset %d, "
				" max: %d\n", 
				offset,
				(SAA_MMAP_ACAP_SZ - 4096));
			return EINVAL;
	}

	pta = &(sc->sc_pta[SAA_CHAN_AUDIO * 1024]);
	*paddr = pta[offset];
	return 0;
}


void
saa_audio_multibuf_switchbuf(struct saa_softc *sc)
{
	int offset;
	struct saa_audio_spec *o;

	o = &sc->sc_audio_spec;
	/* Note increment */
	if (++sc->sc_sts->ssts_audio_mb_cur >=
				sc->sc_sts->ssts_audio_mb_count)
		sc->sc_sts->ssts_audio_mb_cur = 0;

	offset = sc->sc_sts->ssts_audio_framesz * sc->sc_sts->ssts_audio_mb_cur;
	REG_WRITE(
	  (REG_DMA_BASE0(SAA_CHAN_AUDIO) 
			+ (0x4 & (sc->sc_sts->ssts_audio_lastbuf << 2))), 
	  offset);
}
