#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/rman.h>
#include <sys/systm.h>

#include <dev/iicbus/iiconf.h>
#include "iicbus_if.h"

#include <vm/vm.h>
#include <vm/vm_extern.h>
#include <vm/pmap.h>                /* req. by machine/pmap.h       */
#include <machine/pmap.h>           /* vtophys                      */
#include <vm/vm_map.h>
#include <vm/vm_object.h>
#include <vm/vm_kern.h>

#include "saa713x_reg.h"
#include "saa713x_ioctl.h"
#include "saa713x_var.h"
#include "saa713x_i2c.h"

static __inline int
saa_i2c_iserror(uint8_t sts)
{
	switch (sts) {
	case I2C_STS_NODEV:
	case I2C_STS_NOACK:
	case I2C_STS_BUSERR:
	case I2C_STS_ARBLOST:
	case I2C_STS_SEQERR:
	case I2C_STS_STSERR:
			return 1;
	}
	return 0;
}

static __inline int
saa_i2c_isidle(uint8_t sts)
{
	switch (sts) {
	case I2C_STS_IDLE:
	case I2C_STS_DONESTOP:
		return 1;
	}
	return 0;
}

int
saa_i2c_reset(device_t dev,
		u_char speed, u_char addr, u_char *oaddr)
{
	int i;
	uint8_t sts;
	struct saa_softc *sc = (struct saa_softc *)device_get_softc(dev);

restart_reset:
	sts = REG_READB(BREG_I2C_CTL) & I2C_CTL_STS_MASK;
	if (!saa_i2c_iserror(sts)) {
		if ((sts == I2C_STS_DONEWRITE) || (sts == I2C_STS_DONEREAD)) {
			saa_i2c_stop(dev);
			goto restart_reset;
		}
		return 0;
	}
	REG_WRITEB(BREG_I2C_CTL, sts | I2C_ATTR_NOP);
	i = SAA_I2C_TIMEOUT;
	while (i) {
		DELAY(SAA_I2C_DELAY);
		sts = REG_READB(BREG_I2C_CTL);
		if (!saa_i2c_iserror(sts))
			break;
		--i;
	}
	if (!i) {
		device_printf(sc->sc_iicbus_dev, "RESET: timeout "
				"(status 0x%x)\n", (int)sts);
		return ETIMEDOUT;
	}
	return 0; 
}

int
saa_i2c_start(device_t dev, u_char slave, int timeout)
{
	int i;
	uint8_t sts;
	struct saa_softc *sc = (struct saa_softc *)device_get_softc(dev);

	sts = REG_READB(BREG_I2C_CTL) & I2C_CTL_STS_MASK;
	if (!saa_i2c_isidle(sts))
		return EILSEQ;
	REG_WRITEB(BREG_I2C_CTL, sts | I2C_ATTR_START);
	REG_WRITEB(BREG_I2C_DATA, slave);
	i = SAA_I2C_TIMEOUT; 
	while (i) {
		DELAY(SAA_I2C_DELAY);
		sts = REG_READB(BREG_I2C_CTL) & I2C_CTL_STS_MASK;
		if ((sts == I2C_STS_DONEWRITE) || (sts == I2C_STS_DONEREAD))
			break;
		if (saa_i2c_iserror(sts)) {
			device_printf(sc->sc_iicbus_dev, "START: bus error "
					"(status 0x%x loop:%d)\n", (int)sts, i);
			
			return EIO;
		}
		--i;
	}
	if (!i) {
		device_printf(sc->sc_iicbus_dev, "START: timeout "
					"(status 0x%x)\n", (int)sts);
		return ETIMEDOUT;
	}
	return 0;
}

int
saa_i2c_stop(device_t dev)
{
	uint8_t sts;
	int i;
	struct saa_softc *sc = (struct saa_softc *)device_get_softc(dev);

	sts = REG_READB(BREG_I2C_CTL) & I2C_CTL_STS_MASK;
	if ((sts != I2C_STS_DONEWRITE) && (sts != I2C_STS_DONEREAD))
		return EILSEQ;
	REG_WRITEB(BREG_I2C_CTL, sts | I2C_ATTR_STOP);
	i = SAA_I2C_TIMEOUT; 
	while (i) {
		DELAY(SAA_I2C_DELAY);
		sts = REG_READB(BREG_I2C_CTL) & I2C_CTL_STS_MASK;
		if (sts == I2C_STS_DONESTOP)
			break;
		if (saa_i2c_iserror(sts)) {
			device_printf(sc->sc_iicbus_dev, "STOP: bus error "
					"(status 0x%x loop:%d)\n", (int)sts, i);
			
			return EIO;
		}
		--i;
	}
	if (!i) {
		device_printf(sc->sc_iicbus_dev, "STOP: timeout "
				"(status 0x%x)\n", (int)sts);
		return ETIMEDOUT;
	}
	return 0;
}
	
int 
saa_i2c_write(device_t dev, 
	 	char *buf, int len, 
	 	int *bytes, int timeout)
{
	uint8_t sts;
	int i, j;
	struct saa_softc *sc = (struct saa_softc *)device_get_softc(dev);

	sts = REG_READB(BREG_I2C_CTL) & I2C_CTL_STS_MASK;
	if (!(sts & I2C_STS_DONEWRITE))
		return EILSEQ;
	*bytes = j = 0;
	while (len--) {
		REG_WRITEB(BREG_I2C_DATA, buf[j]);
		REG_WRITEB(BREG_I2C_CTL, sts | I2C_ATTR_CONT);
		i = SAA_I2C_TIMEOUT; 
		while (i) {
			DELAY(SAA_I2C_DELAY);
			sts = REG_READB(BREG_I2C_CTL) & I2C_CTL_STS_MASK;
			if (sts == I2C_STS_DONEWRITE)
				break;
			if (saa_i2c_iserror(sts)) {
				device_printf(sc->sc_iicbus_dev, 
					"WRITE: bus error "
					"(status 0x%x loop:%d)\n", (int)sts, i);
			
				return EIO;
			}
			--i;
		}
		if (!i) {
			device_printf(sc->sc_iicbus_dev, "WRITE: timeout "
					"(status 0x%x)\n", (int)sts);
			return ETIMEDOUT;
		}
		++j;
	}
	*bytes = j;
	return 0;
}

int
saa_i2c_read(device_t dev, 
		char *buf, int len, 
		int *bytes, int last, int delay)
{
	uint8_t sts;
	int i, j;
	struct saa_softc *sc = (struct saa_softc *)device_get_softc(dev);

	sts = REG_READB(BREG_I2C_CTL) & I2C_CTL_STS_MASK;
	if (!(sts & I2C_STS_DONEREAD))
		return EILSEQ;
	*bytes = j = 0;
	while (len--) {
		REG_WRITEB(BREG_I2C_CTL, sts | I2C_ATTR_CONT);
		i = SAA_I2C_TIMEOUT; 
		while (i) {
			DELAY(SAA_I2C_DELAY);
			sts = REG_READB(BREG_I2C_CTL) & I2C_CTL_STS_MASK;
			if (sts != I2C_STS_BUSY)
				break;
			if (saa_i2c_iserror(sts)) {
				device_printf(sc->sc_iicbus_dev, 
					"READ: bus error "
					"(status 0x%x loop:%d)\n", (int)sts, i);
				return EIO;
			}
			--i;
		}
		if (!i) {
			device_printf(sc->sc_iicbus_dev, "READ: timeout "
					"(status 0x%x)\n", (int)sts);
			return ETIMEDOUT;
		}
		buf[j] = REG_READB(BREG_I2C_DATA);
		++j;
	}
	*bytes = j;
	return 0;
}
