#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/conf.h>			/* d_open_t, ... */
#include <sys/bus.h>
#include <machine/bus.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/mman.h>
#include <sys/proc.h>
#ifdef SAA_DSHM
#include <sys/shm.h>			/* IPC_STAT, SHM_W */
#endif

#include <vm/vm.h>
#include <vm/vm_extern.h>		/* vm_fault */
#include <vm/pmap.h> 			/* machine/pmap.h */
#include <machine/pmap.h>		/* vm_map_pmap */
#include <vm/vm_map.h>			/* vm_map_find, _unmap, _remove */
#include <vm/vm_object.h>		/* vm_object_reference, _deallocate */
#include <vm/vm_kern.h>			/* kmem_map */
#include <vm/vm_param.h>		/* KERN_SUCCESS */

#include "saa713x_reg.h"
#include "saa713x_ioctl.h"
#include "saa713x_var.h"
#include "saa713x_video.h"

MALLOC_DECLARE(M_SAA_VCAP);
MALLOC_DEFINE(M_SAA_VCAP, "vdocapbuf", "SAA713x Video capture buffer");

#ifdef SAA_DSHM
/* 
 * shm_handle and kern_shmctl decls have to be consistent with
 * sys/kern/sysv_shm.c
 */

struct shm_handle 
{
        vm_object_t shm_object;
};

extern int kern_shmctl(struct thread *td, 
			int shmid, 
			int cmd, 
			void *buf, size_t bufsz);
#endif

static void saa_video_setup_scaler(struct saa_softc *sc, int task);

static void saa_video_setopt(struct saa_softc *sc, struct saa_video_opt *o);
static void saa_video_setstd(struct saa_softc *sc, 
				struct saa_video_std *s,
				int task);

static void saa_video_overlay_off(struct saa_softc *sc);
static int saa_video_overlay_on(struct saa_softc *sc, 
				struct saa_video_spec *o);

static void saa_video_capture_off(struct saa_softc *sc);
static int saa_video_capture_on(struct saa_softc *sc, 
				struct saa_video_spec *o,
				struct thread *td);

#ifdef SAA_DSHM
static void saa_video_dealloc_shm_res(struct saa_softc *sc);
#endif

static int saa_video_capbuf_init(struct saa_softc *sc, int t, int sz);
static int saa_video_capbuf_dnit(struct saa_softc *sc, int t);

static void saa_load_cliplist(struct saa_softc *sc, 
				struct saa_video_clipmode *svcm,
				struct saa_video_cliplist *svcl);

static d_open_t		saa_videodev_open;
static d_close_t	saa_videodev_close;
static d_ioctl_t	saa_videodev_ioctl;
static d_mmap_t		saa_videodev_mmap;

static struct cdevsw saa_video_cdevsw = {
	.d_name 	= "saa",
	.d_open 	= saa_videodev_open,
	.d_close 	= saa_videodev_close,
	.d_ioctl 	= saa_videodev_ioctl,
	.d_mmap		= saa_videodev_mmap,
	DVERSION
	DFLAGS
};

static struct saa_video_std saa_default_video_std = {	/* PAL B,G defaults */
	.std_flag_dfid			= 0,
	.std_flag_no_dved		= 0,
	.std_flag_dhed			= 0,

	.std_first_linenum  		= 0,

	.std_video_xinstart  		= 0,
	.std_video_xinstop 		= 719,
	.std_video_yinstart 		= 19,
	.std_video_yinstop 		= 19+287,

	.std_vbi_xinstart  		= 0,
	.std_vbi_xinstop 		= 0,
	.std_vbi_yinstart		= 0,
	.std_vbi_yinstop		= 0,

	.std_brightness 		= 0x80,
	.std_contrast 			= 0x44,
	.std_saturation 		= 0x40,

	.std_vbi_raw_src 		= VBIRAWSRC_YUV
};

static struct saa_video_opt saa_default_video_opt = { 
	.svo_flag_wpoff			= 0,

	.svo_fuse 			= FUSE_GAON_AAOFF,
	.svo_input_mode 		= MODE_CVBS_2,

	.svo_flag_dtv			= 0,
	.svo_flag_copyprotsrc		= 0,
	.svo_flag_cpoff  		= 0,
	.svo_flag_holdg			= 0,
	.svo_flag_fixgain  		= 0,

	.svo_adc1_gain  		= 0,
	.svo_adc2_gain  		= 0,

	.svo_feild_rate  		= FRATE_AUTO,
	.svo_htc_mode  			= HTCMODE_FAST,
	.svo_flag_nonpll_mode  		= 0,
	.svo_vno_mode  			= VNOMODE_NORM,

	.svo_flag_byps 			= 0,
	.svo_flag_ycomb 		= 1,
	.svo_flag_lubw 			= 1,
	.svo_lufi 			= 8,

	.svo_brightness 		= 0x80,
	.svo_contrast 			= 0x44,
	.svo_saturation 		= 0x40,
	.svo_hue 			= 0x00,

	.svo_flag_no_acgc 		= 0,
	.svo_chroma_gain 		= 0,

	.svo_flag_cdto 			= 0,
	.svo_cstd 			= CSTD_PAL,
	.svo_flag_dcvf 			= 0,
	.svo_flag_fctc 			= 0,
	.svo_flag_ccomb 		= 1,

	.svo_offv 			= 0,
	.svo_offu 			= 0,
	.svo_flag_chbw 			= 1,
	.svo_lcbw 			= 7,

	.svo_ydel			= 0,
	.svo_hdel			= 0,
	.svo_flag_forcecolor		= 0,

	.svo_apck			= 1,
	.svo_aosl			= 0,
	.svo_flag_uptcv			= 1,
	.svo_flag_cm99			= 0,

	.svo_laty			= 3,

	.svo_adetect_std 		= ADETECT_DISABLE,
};


/*
 * Accesable via ioctl SETVIDEOSTD
 * Can be called while capture/overlay is on
 */

static void
saa_video_setstd(struct saa_softc *sc, struct saa_video_std *s, int task)
{
	uint8_t v, i;

	v = (s->std_first_linenum << SRC_TMNG1_YPOS_LOBITS_SHIFT)
						& SRC_TMNG1_YPOS_LOBITS_MASK;
	REG_WRITEB(BREG_SRC_TMNG1, v);

	v = ((s->std_first_linenum >> 8) << SRC_TMNG2_YPOS_HIBITS_SHIFT)
						& SRC_TMNG2_YPOS_HIBITS_MASK;

	/* Note dved flag is reversed for easy default specification */
	if (s->std_flag_no_dved == 0)
		v |= SRC_TMNG2_DVED;
	if (s->std_flag_dfid != 0)
		v |= SRC_TMNG2_DFID;
	if (s->std_flag_dhed != 0)
		v |= SRC_TMNG2_DHED;

	REG_WRITEB(BREG_SRC_TMNG2,  v);
	
	REG_WRITEW(WREG_TASK_VIDEO_XINSTART(task), s->std_video_xinstart);
	REG_WRITEW(WREG_TASK_VIDEO_XINSTOP(task), s->std_video_xinstop);
	REG_WRITEW(WREG_TASK_VIDEO_YINSTART(task), s->std_video_yinstart);
	REG_WRITEW(WREG_TASK_VIDEO_YINSTOP(task), s->std_video_yinstop);

	REG_WRITEW(WREG_TASK_VBI_XINSTART(task), s->std_vbi_xinstart);
	REG_WRITEW(WREG_TASK_VBI_XINSTOP(task), s->std_vbi_xinstop);
	REG_WRITEW(WREG_TASK_VBI_YINSTART(task), s->std_vbi_yinstart);
	REG_WRITEW(WREG_TASK_VBI_YINSTOP(task), s->std_vbi_yinstop);

	REG_WRITEB(BREG_TASK_LUMA_BRIGHT(task), s->std_brightness);
	REG_WRITEB(BREG_TASK_LUMA_CONT(task), s->std_contrast);
	REG_WRITEB(BREG_TASK_CHROMA_SATN(task), s->std_saturation);

	/* This rigmarole to avoid overwriting CSCON, MACON */
	i = REG_READB(BREG_TASK_DPATH(task));
	i &= ~TASK_DPATH_VBI_RAWS_MASK;
	i |= s->std_vbi_raw_src;
	REG_WRITEB(BREG_TASK_DPATH(task), i);
}


/*
 * Accesable via ioctl SETVIDEOOPT
 * Can be called while capture/overlay is on
 * For auto color standard detection and deciding between NTSC, PAL, etc.
 * the number of frames to check before deciding on a standard is
 * set by LATY. For auto detection FCTC must be 0.
 * CSTD sets the preferred standard.
 * ADETECT_LEVEL3:
 *	Hardware sets LUBW, LCBW, CHBW, DCVF and LUF based on input signal
 * ADETECT_LEVEL2:
 *	Hardware sets LUBW, LCBW, CHBW, DCVF based on input signal
 * ADETECT_LEVEL1:
 *	Hardware sets DCVF
 */

static void
saa_video_setopt(struct saa_softc *sc, struct saa_video_opt *o)
{
	uint8_t v;

	v = DEFAULT_INCR_DELAY;
	if (o->svo_flag_wpoff || o->svo_flag_dtv)
		v |= INCR_DELAY_WPOFF;
	REG_WRITEB(BREG_INCR_DELAY, v);

	REG_WRITEB(BREG_AINPUT1, (o->svo_fuse | o->svo_input_mode));

	v = o->svo_flag_dtv ? AINPUT2_HLNRS : 0;
	v |= o->svo_flag_copyprotsrc ? AINPUT2_VBSL : 0;
	v |= o->svo_flag_cpoff ? AINPUT2_CPOFF : 0;
	v |= o->svo_flag_holdg ? AINPUT2_HOLDG : 0;
	v |= o->svo_flag_fixgain ? AINPUT2_GAINFIX : 0;
	v |= o->svo_adc1_gain & 0x100 ? AINPUT2_GAIN1_BIT8 : 0;
	v |= o->svo_adc2_gain & 0x100 ? AINPUT2_GAIN2_BIT8 : 0;
	REG_WRITEB(BREG_AINPUT2, v);

	REG_WRITEB(BREG_AINPUT3_GAIN1, (uint8_t)(o->svo_adc1_gain & 0xff));
	REG_WRITEB(BREG_AINPUT4_GAIN2, (uint8_t)(o->svo_adc2_gain & 0xff));
	REG_WRITEB(BREG_HGB, DEFAULT_HGB);
	REG_WRITEB(BREG_HGE, DEFAULT_HGE);

	if (o->svo_feild_rate == 0)
		v = SYNC_CTRL_AUFD;
	else if (o->svo_feild_rate == FRATE_60)
		v = SYNC_CTRL_FSEL;
	v |= o->svo_htc_mode;
	v |= o->svo_flag_nonpll_mode ? SYNC_CTRL_HPLL : 0;
	v |= o->svo_vno_mode;
	REG_WRITEB(BREG_SYNC_CTRL, v);

	v = o->svo_flag_byps ? LUMA_CTRL_BYPS : 0;
	v |= o->svo_flag_ycomb ? LUMA_CTRL_YCOMB : 0;
	v |= o->svo_flag_lubw ? LUMA_CTRL_LUBW : 0;
	v |= (o->svo_lufi << LUMA_CTRL_LUFI_SHIFT) & LUMA_CTRL_LUFI_MASK;
	REG_WRITEB(BREG_LUMA_CTRL, v);

	REG_WRITEB(BREG_LUMA_BRIGHT, o->svo_brightness);
	REG_WRITEB(BREG_LUMA_CONT, o->svo_contrast);
	REG_WRITEB(BREG_CHROMA_SATN, o->svo_saturation);
	REG_WRITEB(BREG_CHROMA_HUE, o->svo_hue);

	v = o->svo_flag_cdto ? CHROMA_CTRL1_CDTO : 0;
	v |= o->svo_cstd;
	v |= o->svo_flag_dcvf ? CHROMA_CTRL1_DCVF : 0;
	v |= o->svo_flag_fctc && 
			o->svo_adetect_std == 0 ? CHROMA_CTRL1_FCTC : 0;
	/* ccomb required when ycomb is active */
	v |= (o->svo_flag_ccomb | o->svo_flag_ycomb)  ? 
					CHROMA_CTRL1_CCOMB : 0;
	REG_WRITEB(BREG_CHROMA_CTRL1, v);

	v = o->svo_flag_no_acgc ? CHROMA_GAIN_ACGC : 0;
	v |= (o->svo_chroma_gain << CHROMA_GAIN_SHIFT) & CHROMA_GAIN_MASK;
	REG_WRITEB(BREG_CHROMA_GAIN, v);

	v = (o->svo_offu  << CHROMA_CTRL2_OFFU_SHIFT) & CHROMA_CTRL2_OFFU_MASK;
	v |= (o->svo_offv  << CHROMA_CTRL2_OFFV_SHIFT) & CHROMA_CTRL2_OFFV_MASK;
	v |= o->svo_flag_chbw ? CHROMA_CTRL2_CHBW : 0;
	v |= (o->svo_lcbw << CHROMA_CTRL2_LCBW_SHIFT) & CHROMA_CTRL2_LCBW_MASK;
	REG_WRITEB(BREG_CHROMA_CTRL2, v);

	v = (o->svo_flag_forcecolor ? DELAY_CTRL_COLO : 0);
	v |= (o->svo_ydel << DELAY_CTRL_YDEL_SHIFT) & DELAY_CTRL_YDEL_MASK; 
	v |= (o->svo_hdel << DELAY_CTRL_HDEL_SHIFT) & DELAY_CTRL_HDEL_MASK; 
	REG_WRITEB(BREG_DELAY_CTRL, v);

	/* XXX May need UPTCV (gain correction per line/feild) */
	v = (o->svo_flag_cm99 ? ADC_CTRL_CM99 : 0);
	v |= (o->svo_flag_uptcv ? ADC_CTRL_UPTCV : 0);
	v |= (o->svo_aosl << ADC_CTRL_AOSL_SHIFT) & ADC_CTRL_AOSL_MASK;
	v |= (o->svo_apck << ADC_CTRL_APCK_SHIFT) & ADC_CTRL_APCK_MASK;
	REG_WRITEB(BREG_ADC_CTRL, v);

	REG_WRITEB(BREG_VGB, DEFAULT_VGB);
	REG_WRITEB(BREG_VGE, DEFAULT_VGE);

	v = (o->svo_laty << MISC_CTRL_LATY_SHIFT) & MISC_CTRL_LATY_MASK;
	REG_WRITEB(BREG_MISC_CTRL, v);

	REG_WRITEB(BREG_VBI_GAIN, DEFAULT_VBI_GAIN);
	REG_WRITEB(BREG_VBI_OFFSET, DEFAULT_VBI_OFFSET);

	if (o->svo_adetect_std & 0x01)
		REG_SETB(BREG_CHROMA_CTRL1, CHROMA_CTRL1_AUTO0);
	else
		REG_CLRB(BREG_CHROMA_CTRL1, CHROMA_CTRL1_AUTO0);
	if (o->svo_adetect_std & 0x10)
		REG_SETB(BREG_ADC_CTRL, ADC_CTRL_AUTO1);
	else
		REG_CLRB(BREG_ADC_CTRL, ADC_CTRL_AUTO1);
}

#ifdef SAA_DSHM
/*
 * Helper function used to dealloc DSHM mapping allocated previously
 * in saa_video_capture_on 
 */

static void
saa_video_dealloc_shm_res(struct saa_softc *sc)
{
	int t;
	for (t = 0; t < 2; ++t) {
		if (sc->sc_video_capbuf_shmaddr[t]) {
			vm_map_unwire(kmem_map,
					sc->sc_video_capbuf_shmaddr[t],
					sc->sc_video_capbuf_shmaddr[t] +
						sc->sc_video_capbuf_shmlen[t],
					0);
			vm_map_remove(kmem_map,
					sc->sc_video_capbuf_shmaddr[t],
					sc->sc_video_capbuf_shmaddr[t] +
					        sc->sc_video_capbuf_shmlen[t]);
			sc->sc_video_capbuf_shmaddr[t] = 0;
		}

		if (sc->sc_video_capbuf_shmobj[t]) {
			vm_object_deallocate(
				sc->sc_video_capbuf_shmobj[t]);
			sc->sc_video_capbuf_shmobj[t] = 0;
		}
	}
}
#endif


/*
 * Terminates video capture and releases resources
 */

static void
saa_video_capture_off(struct saa_softc *sc)
{
	REG_CLR(REG_IEN, sc->sc_video_spec.svs_cap_sigmask);
	REG_CLRB(REG_FIEN, (FIEN_SLICE0_R0_T0_OE1 | 
				FIEN_SLICE0_R0_T1_OE1 |
				FIEN_SLICE0_R0_T0_OE0 |
				FIEN_SLICE0_R0_T1_OE0));
	REG_CLRB(BREG_REGION_EN, (REGION_EN_VIDEO_A | REGION_EN_VIDEO_B));
	REG_CLR(REG_MCR, (MCR_CHAN0_TE | MCR_CHAN1_TE | 
				MCR_CHAN4_TE | MCR_CHAN5_TE));
	sc->sc_video_flags &= ~(SCVF_ON | SCVF_CAPTURE | SCVF_DMA_PAUSED);
#ifdef SAA_DSHM
	saa_video_dealloc_shm_res(sc);
#endif
}


/*
 * Main video capture function, 
 * Video capture modes possible:
 *		DSHM - Planar - Task A
 *		DSHM - Planar - Task A w/ Multibuf
 *		DSHM - Packed - Task A
 *		DSHM - Packed - Task B
 *		DSHM - Packed - Task A w/ Multibuf
 *		DSHM - Packed - Task B w/ Multibuf
 *		DSHM - Packed - Task A + Task B
 *		DSHM - Packed - Task A w/ Multibuf + Task B w/ Multibuf
 *
 *		CBUF - Planar - Task A
 *		CBUF - Planar - Task A w/ Multibuf
 *		CBUF - Packed - Task A
 *		CBUF - Packed - Task B
 *		CBUF - Packed - Task A w/ Multibuf
 *		CBUF - Packed - Task B w/ Multibuf
 *		CBUF - Packed - Task A + Task B
 *		CBUF - Packed - Task A w/ Multibuf + Task B w/ Multibuf
 *
 * About task A/B: in packed modes the client can run two parallel
 * tasks. These tasks are executed one after the other
 * each processing video_frame_rate / 2, this behaviour
 * can be altered using svs_cap_frameskip[task_nr]. Frame
 * skip works for single tasks also.
 *
 * DSHM is custom extension which allows clients to use 
 * SYSV shared memory for capture. The device DMA's directly
 * to the shared memory pages. The shared memory id's passed
 * by the clients are first validated and then mapped into 
 * kernel address space. After this all shm pages are made
 * resident and marked wired.
 *
 * Multibuf is a custom extension which allows clients to specify
 * an oversized area for capture (one which will fit 1 + n frames)
 * the driver automatically calculates the max. frames that will
 * fit in the buffer and sequentially steps through the areas.
 * This increases the latency limit of the client. For e.g
 * in I420 mode using a 4MB buffer we can hold six frames instead
 * of one. Clients mmap the STS page to access information
 * pertaining to the currently active area, task etc.
 */ 

static int
saa_video_capture_on(struct saa_softc *sc, struct saa_video_spec *o,
			struct thread *td)
{
#ifdef SAA_DSHM
	struct shmid_ds shm_ds[2];
	vm_offset_t vo;			
#endif
	uint32_t *ptr[2][3];
	int t, idx, l, ctrl;		/* General purpose vars */
	uint8_t bpp;			/* Bits per pixel */

	bpp = (o->svs_vfmt & VFMT_BPP_MASK) >> VFMT_BPP_SHIFT;
	l = (o->svs_buf_width  * o->svs_buf_height * bpp) / 8;

	/* 
	 * Picture dimensions too big 
	 */
	if (l > SAA_VCAP_FRAME_MAX_SZ ||
	    (o->svs_flags & VSPEC_FLAG_CAPTURE_MULTIBUF &&
		o->svs_cap_mbufsz > 0 && l > o->svs_cap_mbufsz))
		return ENOSPC;

	/* 
	 * Planar mode supports one task (TaskA) only 
	 */
	if (o->svs_cap_nrtasks < 1 || 
	    o->svs_cap_nrtasks > 2 ||
	    ((o->svs_cap_nrtasks == 2) && (o->svs_vfmt & VFMT_PLANAR)))
		return EINVAL;

	/* We setup ptr for convienience as planar, nonplanar,
	 * shm, nonshm code all need to access the ptas
	 * furthermore we bzero the pta to avoid memory 
	 * corruption from existing entires set during last 
	 * capture
         */
	for (t = 0; t < 2; ++t) {
		bzero(&(sc->sc_pta[t * 1024]), 4096);
		ptr[t][0] = &(sc->sc_pta[t * 1024 ]);

		bzero(&(sc->sc_pta[SAA_CHAN_YUV_U * 1024]), 4096);
		ptr[t][1] = &(sc->sc_pta[SAA_CHAN_YUV_U * 1024 ]);

		bzero(&(sc->sc_pta[SAA_CHAN_YUV_V * 1024]), 4096);
		ptr[t][2] = &(sc->sc_pta[SAA_CHAN_YUV_V * 1024 ]);
	}

	/*
	 * Client request capture into a malloced buffer, owned
 	 * by the driver. The client will mmap this buffer into
	 * its own address space via videodev_mmap
	 * This facility is exclusive WRT DSHM (meaning both
	 * cannot be used simultaneously)
	 * 
	 * We mark allocflag in a separate loop to avoid
	 * unmarking, which we would have to do if we
	 * were using one loop only and encountered a
	 * failed capbuf_init return
	 * By marking we also tell videodev_mmap what buffers,
	 * if any are available for mmap
	 *
	 * capbuf_init, capbuf_dnit do the hardwork ...
	 */
	if ((o->svs_flags & VSPEC_FLAG_CAPTURE_SHM) == 0) {
		if (l > o->svs_cap_bufsz)
			return ENOSPC;

		for (t = 0; t < o->svs_cap_nrtasks; ++t) {
			l = saa_video_capbuf_init(sc, t, o->svs_cap_bufsz);
			if (l != 0) {
		l1_redo_dnit:
				idx = saa_video_capbuf_dnit(sc, t);
				if (idx != 0)
					device_printf(sc->sc_dev,
							"capbuf deallocation "
							"failed. Mem. leak\n");
				if (t == 1) {
					--t;
					goto l1_redo_dnit;
				}
				return (idx == 0 ? ENOMEM : idx);
			}
		}

		/* Mark allocated buffers for videodev_mmap */
		for (t = 0; t < 2; ++t) {
			if (t < o->svs_cap_nrtasks)
				sc->sc_video_capbuf_allocflag[t] = 1;
			else
				sc->sc_video_capbuf_allocflag[t] = 0;
		}
	}

	/*
	 * Client requests DSHM capabilty, validate passed
	 * shm_ids, permission and map pages into sc_pta
	 *
	 * Connect to the SHM segment(s) in question
	 * Map into in to kmem_map
	 * Fault in pages using vm_fault
	 * wire faulted in pages
	 * use vtophsy to obtain phys. addrs
	 * build pta(s)
	 */

	/* XXX Move to stand alone function and shift to
	 * saa713x.c as audio module will require the
	 * same functionality
	 */

	if (o->svs_flags & VSPEC_FLAG_CAPTURE_SHM) {
#ifdef SAA_DSHM
		for (t = 0; t < o->svs_cap_nrtasks; ++t) {
			if (o->svs_cap_shmid[t] == 0) {
				saa_video_dealloc_shm_res(sc);
				return EINVAL;
			}

			if (kern_shmctl(td, o->svs_cap_shmid[t], 
					IPC_STAT, &shm_ds[t], 0) != 0) {
				saa_video_dealloc_shm_res(sc);
				return ENOENT;
			}

			/* 
			 * Direct Semapore capability limited to
			 * super user or semaphore creator, 
			 * additionally SHM should have
			 * at least one write bit set 
			 */
			l = SHM_W | (SHM_W >> 3) | (SHM_W >> 6);
			if ((shm_ds[t].shm_perm.mode & l) == 0 || 
			    (!suser(td) && 
			     shm_ds[t].shm_perm.cuid != td->td_ucred->cr_uid)
			   ) {
				saa_video_dealloc_shm_res(sc);
				return EPERM;
			}

			if (((o->svs_buf_width * o->svs_buf_height * bpp) / 8)
							> shm_ds[t].shm_segsz) {
				saa_video_dealloc_shm_res(sc);
				return ENOSPC;
			}

			/* 
			 * In Multibuf mode we require the shared memory
			 * segment to be of size >= SAA_VCAMP_FRAME_MAX_SZ
			 */
			if (o->svs_flags & VSPEC_FLAG_CAPTURE_MULTIBUF &&
				shm_ds[t].shm_segsz < SAA_VCAP_FRAME_MAX_SZ) {
				saa_video_dealloc_shm_res(sc);
				return ENOSPC;
			}

			sc->sc_video_capbuf_shmobj[t] = shm_ds[t].shm_internal;
			if (sc->sc_video_capbuf_shmobj[t] == 0) {
				saa_video_dealloc_shm_res(sc);
				return EINVAL;
			}

			/* Refrence and map shm object into kernel space */
			vm_object_reference(sc->sc_video_capbuf_shmobj[t]);

			sc->sc_video_capbuf_shmaddr[t] = 0;
			sc->sc_video_capbuf_shmlen[t] = shm_ds[t].shm_segsz;
			if (vm_map_find(kmem_map, 
				sc->sc_video_capbuf_shmobj[t], 
					0,
					&sc->sc_video_capbuf_shmaddr[t],
					shm_ds[t].shm_segsz,
					1,
					VM_PROT_READ | VM_PROT_WRITE, 
					VM_PROT_READ | VM_PROT_WRITE, 
					MAP_PREFAULT) != KERN_SUCCESS) {
				sc->sc_video_capbuf_shmaddr[t] = 0;
				saa_video_dealloc_shm_res(sc);
				return EINVAL;
			}
		
			/* Fault in lazy pages */
			idx = l = 0;
			while(l < sc->sc_video_capbuf_shmlen[t] &&
		      		idx < 1024) {
				vo = sc->sc_video_capbuf_shmaddr[t] + l;
		  	    	ptr[t][0][idx] = vtophys(vo);
				if (ptr[t][0][idx] == 0) {
					vm_fault(kmem_map,
						vo,
						VM_PROT_WRITE,
						0);
		  	    		ptr[t][0][idx] = vtophys(vo);
					if (ptr[t][0][idx] == 0) {
						saa_video_dealloc_shm_res(sc);
						return ENOMEM;
					}
				}
		  		ptr[t][1][idx] =  ptr[t][0][idx];
		  		ptr[t][2][idx] =  ptr[t][0][idx];
				++idx;
				l += 4096;
			}	

			/* Wire pages */
			if (vm_map_wire(kmem_map, 
				sc->sc_video_capbuf_shmaddr[t],
				sc->sc_video_capbuf_shmaddr[t] + 
				  sc->sc_video_capbuf_shmlen[t],
				0) != KERN_SUCCESS) {
				saa_video_dealloc_shm_res(sc);
				return ENOMEM;
			}
		} /* end for(0->nrtasks) */
#else 	/*#ifdef SAA_DSHM*/
		return ENOTSUP;
#endif
	} /* end FLAG_SHM */

	/* No failure returns beyond this point */

	saa_video_setopt(sc, &sc->sc_video_opt);
	for (t = 0; t < o->svs_cap_nrtasks; ++t) {
		saa_video_setstd(sc, &sc->sc_video_std, t);

		if (o->svs_flags & VSPEC_FLAG_CAPTURE_EVENFEILD_FIRST) {
			REG_WRITEB(BREG_TASK_COND(t), 
				(o->svs_flags & VSPEC_FLAG_FRAMELOCK ? 
							TASK_COND_SLCK : 0) |
			       TASK_COND_U50 | TASK_COND_U60 | TASK_COND_UEVEN);
		} else {
			REG_WRITEB(BREG_TASK_COND(t), 
				(o->svs_flags & VSPEC_FLAG_FRAMELOCK ? 
							TASK_COND_SLCK : 0) |
			       TASK_COND_U50 | TASK_COND_U60 | TASK_COND_UODD);
		}

		if (o->svs_flags & VSPEC_FLAG_CAPTURE_ONEFEILD) {
			REG_WRITEB(BREG_TASK_FEILD(t),
			 ((o->svs_cap_frameskip[t] << TASK_FEILD_FSKIP_SHIFT) & 
						TASK_FEILD_FSKIP_MASK) |  
			 ((1 <<  TASK_FEILD_COUNT_SHIFT) & 
						TASK_FEILD_COUNT_MASK));
		} else {
			REG_WRITEB(BREG_TASK_FEILD(t),
			 ((o->svs_cap_frameskip[t] << TASK_FEILD_FSKIP_SHIFT) & 
						TASK_FEILD_FSKIP_MASK) |  
			 ((2 <<  TASK_FEILD_COUNT_SHIFT) & 
						TASK_FEILD_COUNT_MASK));
		}

		/* CSCON - RGB to YUV filter */
		/* We use CLRB, SETB, to avoid overwriting VBIRAWSRC */
		if ((o->svs_vfmt & VFMT_YUV) == 0) {
			REG_CLRB(BREG_TASK_DPATH(t), TASK_DPATH_CSCON);
			REG_SETB(BREG_TASK_DPATH(t), TASK_DPATH_MACON);
		} else {
		/* 
		 * instead of CLRB(...) we could use 
		 * SETB(TASK_DPATH_MACON | TASK_DPATH_CSCON)
		 * which is a must when we use GACON (XXX GACON not impl.)
		 */
			REG_CLRB(BREG_TASK_DPATH(t), 
					(TASK_DPATH_CSCON | TASK_DPATH_MACON));
		}

		REG_WRITEW(WREG_TASK_VIDEO_XOUTLEN(t), o->svs_buf_width);
		REG_WRITEW(WREG_TASK_VIDEO_YOUTLEN(t), o->svs_buf_height);

		REG_WRITEB(BREG_TASK_VIDEO_FMT(t), 
					(o->svs_vfmt & VFMT_FMT_MASK) |
					((o->svs_flags & VSPEC_FLAG_DITHER) 
								? 0x80 : 0));
		/* XXX Will reset INTLCD and CLINV */
		REG_WRITEB(BREG_VIDEO_UVL, 
			((o->svs_vfmt & VFMT_UVP_MASK) >> VFMT_UVP_SHIFT) &
			VIDEO_UVL_MASK);

		saa_video_setup_scaler(sc, t);

		if (o->svs_vfmt & VFMT_PLANAR) {
			REG_WRITE(REG_DMA_BASE0(SAA_CHAN_YUV_Y),
				(int)o->svs_buf_base + o->svs_buf_offset[0][0]);
			REG_WRITE(REG_DMA_BASE1(SAA_CHAN_YUV_Y),
				(int)o->svs_buf_base + o->svs_buf_offset[1][0]);
			REG_WRITE(REG_DMA_PITCH(SAA_CHAN_YUV_Y),
							o->svs_buf_pitch[0]);

			REG_WRITE(REG_DMA_BASE0(SAA_CHAN_YUV_U), 
				(int)o->svs_buf_base + o->svs_buf_offset[0][1]);
			REG_WRITE(REG_DMA_BASE1(SAA_CHAN_YUV_U), 
				(int)o->svs_buf_base + o->svs_buf_offset[1][1]);
			REG_WRITE(REG_DMA_PITCH(SAA_CHAN_YUV_U),
						o->svs_buf_pitch[1]);

			REG_WRITE(REG_DMA_BASE0(SAA_CHAN_YUV_V), 
				(int)o->svs_buf_base + o->svs_buf_offset[0][2]);
			REG_WRITE(REG_DMA_BASE1(SAA_CHAN_YUV_V), 
				(int)o->svs_buf_base + o->svs_buf_offset[1][2]);
			REG_WRITE(REG_DMA_PITCH(SAA_CHAN_YUV_V),
						o->svs_buf_pitch[2]);

			ctrl = 	(0x7 << DMA_CTRL_BURST_SHIFT) 
						& DMA_CTRL_BURST_MASK;
			ctrl |= DMA_CTRL_ME;
			ctrl |= o->svs_cap_swapbytes;

			REG_WRITE(REG_DMA_CTRL(SAA_CHAN_YUV_Y),
			    ctrl | ((sc->sc_pta_paddr[SAA_CHAN_YUV_Y] >> 12) 
							  & DMA_CTRL_PTA_MASK));
			REG_WRITE(REG_DMA_CTRL(SAA_CHAN_YUV_U),
			    ctrl | ((sc->sc_pta_paddr[SAA_CHAN_YUV_U] >> 12)
							  & DMA_CTRL_PTA_MASK));
			REG_WRITE(REG_DMA_CTRL(SAA_CHAN_YUV_V),
			    ctrl | ((sc->sc_pta_paddr[SAA_CHAN_YUV_V] >> 12)
						 	  & DMA_CTRL_PTA_MASK));
		} else {
			REG_WRITE(REG_DMA_BASE0(t), 
	       			(int)o->svs_buf_base + o->svs_buf_offset[0][0]);
			REG_WRITE(REG_DMA_BASE1(t),
				(int)o->svs_buf_base + o->svs_buf_offset[1][0]);
			REG_WRITE(REG_DMA_PITCH(t), o->svs_buf_pitch[0]);
			ctrl = 	(0x7 << DMA_CTRL_BURST_SHIFT) 
						& DMA_CTRL_BURST_MASK;
			ctrl |= DMA_CTRL_ME;
			ctrl |= o->svs_cap_swapbytes;
			ctrl |= ((sc->sc_pta_paddr[t] >> 12)
							& DMA_CTRL_PTA_MASK);
			REG_WRITE(REG_DMA_CTRL(t), ctrl);
		}
	}

	l = REGION_EN_VIDEO_A;
	ctrl = MCR_CHAN0_TE;
	if (o->svs_cap_nrtasks == 2) {
		l |= REGION_EN_VIDEO_B;
		ctrl |= MCR_CHAN1_TE;
	}

	if (o->svs_vfmt & VFMT_PLANAR) {
		FIFO_SET_SIZE(384, 256, 256, 128);
		ctrl |=  MCR_CHAN4_TE  | MCR_CHAN5_TE;
	} else 
		FIFO_SET_SIZE(896, 0, 0, 128);

	sc->sc_sts->ssts_video_sig_count = 0;
	sc->sc_sts->ssts_video_framectr = 0;
	sc->sc_sts->ssts_video_lasttask = 0;
	sc->sc_sts->ssts_video_t1_framectr = 0;
	sc->sc_sts->ssts_video_t2_framectr = 0;
	sc->sc_sts->ssts_video_feild = 1;	/* Default for NON SIGFLD */

	/* rounded page size */
	sc->sc_sts->ssts_video_framesz = 
			((o->svs_buf_width * o->svs_buf_height * bpp) + 7) / 8;

	/* Multi Buffer */
	sc->sc_sts->ssts_video_mb_cur[0] = 0;	/* Last for task A */
	sc->sc_sts->ssts_video_mb_cur[1] = 0;	/* Last for task B */
	sc->sc_sts->ssts_video_mb_count = 1;
	sc->sc_sts->ssts_video_mbufsz = 0;
	if (o->svs_flags & VSPEC_FLAG_CAPTURE_MULTIBUF) {
		if (o->svs_cap_mbufsz == 0) {
			sc->sc_sts->ssts_video_mbufsz = 
					sc->sc_sts->ssts_video_framesz;
		} else {
			sc->sc_sts->ssts_video_mbufsz = o->svs_cap_mbufsz;
		}
		sc->sc_sts->ssts_video_mb_count = SAA_VCAP_FRAME_MAX_SZ / 
						sc->sc_sts->ssts_video_mbufsz;
		if ((sc->sc_sts->ssts_video_mb_count > 
						o->svs_cap_mb_limit) &&
		    (o->svs_cap_mb_limit > 0))
		    sc->sc_sts->ssts_video_mb_count = o->svs_cap_mb_limit;
	}

	saa_load_cliplist(sc, &o->svs_clipmode, &o->svs_cliplist);

	/* Take off :) */
	REG_WRITEB(BREG_REGION_EN, l);
	REG_SET(REG_MCR, ctrl);			/* Note use of SET not WRITE */

	/* 
	 * End of feild 2 interrupts are always enabled, ie.,
	 * even when VSPEC_FLAG_CAPTURE_SIGNAL is not specified as
	 * we always need to update status (frame_ctr, multibuf, etc.)
	 * when capture is on.
	 * If VSPEC_FLAG_CAPUTRE_SIGNAL is not saa_intr will not deliver
	 * video signals.
	 */
	l = FIEN_SLICE0_R0_T0_OE1 | FIEN_SLICE0_R0_T1_OE1;
	if (o->svs_flags & VSPEC_FLAG_CAPTURE_SIGFLD)
		l |= FIEN_SLICE0_R0_T0_OE0 | FIEN_SLICE0_R0_T1_OE0;
	REG_SET(REG_FIEN, l);			/* Note use of SET not WRITE */

	if (o->svs_flags & VSPEC_FLAG_STATUS_SIGNAL) {
		o->svs_cap_sigmask &= STS_ALL;
		REG_SET(REG_IEN, o->svs_cap_sigmask);
	} else {
		o->svs_cap_sigmask = 0;
		REG_CLR(REG_IEN, STS_ALL);
	}

	sc->sc_video_flags |= SCVF_ON | SCVF_CAPTURE;
	return 0;
}


static void
video_dma_cb(void *arg, bus_dma_segment_t *segs, int nsegs, int error)
{
	struct saa_softc *sc = (struct saa_softc*)arg;
	int i, t;
	
	if (error == EFBIG) {
		sc->sc_capbuf_loaderror = EFBIG;
		wakeup(sc);
		return;
	}	

	/* Invariant loadidx < 1024 */
	t = sc->sc_video_capbuf_current;
	i = 0;
	while (i < nsegs) {
	   sc->sc_capbuf_pgsegs[t][sc->sc_capbuf_loadidx++] =
				segs[i];
	   sc->sc_capbuf_loadlen += segs[i++].ds_len;
	}
	sc->sc_video_capbuf_loadcount[t] += nsegs;

	if (sc->sc_capbuf_loadlen >= sc->sc_video_capbuf_sz[t]) 
		wakeup(sc);
}


static int
saa_video_capbuf_init(struct saa_softc *sc, int t, int sz)
{
	int rval;

	/* max size 4 MB */
	if (sz != SAA_VCAP_FRAME_MAX_SZ)
		return EINVAL;

	if (sc->sc_video_capbuf[t] == NULL) {
		sc->sc_video_capbuf_current = t;
		sz = SAA_VCAP_FRAME_MAX_SZ;
		sc->sc_video_capbuf_sz[t] = sz;
		sc->sc_video_capbuf[t] = malloc(sz, M_SAA_VCAP, 
							M_WAITOK | M_ZERO);
		if (sc->sc_video_capbuf[t] == NULL) {
			device_printf(sc->sc_dev, "Error malloc'ing capture "
				  		"buffer nr. %d\n", t);
			return ENOMEM;
		}
		if (bus_dma_tag_create(0, 
				4096, 0,
				BUS_SPACE_MAXADDR_32BIT,
				BUS_SPACE_MAXADDR,
				0, 0,
				sz, 1024, sz,	/* Device's MMU Pagetable */
				0,		/* has 1024 entries */
				0, 0,
				&sc->sc_video_capbuf_tag[t]) != 0) {
			device_printf(sc->sc_dev, "Error creating capture "
						"buffer tag nr. %d\n", t);
			return EAGAIN;
		}
		if (bus_dmamap_create(sc->sc_video_capbuf_tag[t],
					0,
					&sc->sc_video_capbuf_map[t]) != 0) {
			device_printf(sc->sc_dev, "Error creating capture "
						"buffer map nr. %d\n", t);
			return EAGAIN;
		}
		sc->sc_video_capbuf_map_destroyflag[t] = 1;
		sc->sc_video_capbuf_loadcount[t] = 0;
		sc->sc_capbuf_loadlen = 0;
		sc->sc_capbuf_loadidx = 0;
		sc->sc_capbuf_loaderror = 0;
		rval = bus_dmamap_load(sc->sc_video_capbuf_tag[t],
					sc->sc_video_capbuf_map[t],
					sc->sc_video_capbuf[t], sz,
					video_dma_cb, sc,
					0);
		if (rval == EINPROGRESS) {
			tsleep(sc, PRI_MAX_KERN, "dmapld", 0);
		} else if (rval != 0) {
			device_printf(sc->sc_dev, "Error loading capture "
					 	"buffer nr. %d\n", t);
			return rval;
		}

		if (sc->sc_capbuf_loaderror != 0) {
			device_printf(sc->sc_dev, "Error, callback failed "
					"to load capture buffer "
					"nr %d (idx: %d)\n", 
					t,
					sc->sc_video_capbuf_loadcount[t]);
			return EAGAIN;
		}
		sc->sc_video_capbuf_map_unloadflag[t] = 1;
	}

	saa_build_pta(sc, 
			sc->sc_capbuf_pgsegs[t], 
			sc->sc_video_capbuf_loadcount[t],
			&(sc->sc_pta[t * 1024]));
	memcpy(&(sc->sc_pta[SAA_CHAN_YUV_U * 1024]),
		&(sc->sc_pta[t * 1024]), 4096);
	memcpy(&(sc->sc_pta[SAA_CHAN_YUV_V * 1024]),
		&(sc->sc_pta[t * 1024]), 4096);
	return 0;
}


static int
saa_video_capbuf_dnit(struct saa_softc *sc, int t)
{
	int rval = 0;

	/* flags required as capbuf_map can legally be equal
	 * to zero
	 */
	if (sc->sc_video_capbuf_map_unloadflag[t]) 
		bus_dmamap_unload(sc->sc_video_capbuf_tag[t],
					sc->sc_video_capbuf_map[t]);
	
	if (sc->sc_video_capbuf_map_destroyflag[t])
		rval = bus_dmamap_destroy(
					sc->sc_video_capbuf_tag[t],
					sc->sc_video_capbuf_map[t]);	

	if (!rval && sc->sc_video_capbuf[t]) {
		free(sc->sc_video_capbuf[t], M_SAA_VCAP);
		sc->sc_video_capbuf[t] = 0;
	}

	if (!rval && sc->sc_video_capbuf_tag[t])
	    rval = bus_dma_tag_destroy(sc->sc_video_capbuf_tag[t]);

	return rval;
}


int
saa_video_init(struct saa_softc *sc)
{
	int unit;

	unit = device_get_unit(sc->sc_dev);
	sc->sc_video_opt = saa_default_video_opt;
	sc->sc_video_std = saa_default_video_std; 
	saa_video_setopt(sc, &saa_default_video_opt);
	saa_video_setstd(sc, &saa_default_video_std, 0);
	saa_video_setstd(sc, &saa_default_video_std, 1);

	/* Invariat: sizeof(struct saa_sts) < 4096 */
	sc->sc_sts = (struct saa_status*) malloc(4096, M_SAA_VCAP, 
							M_WAITOK | M_ZERO);
	if (sc->sc_sts == NULL) {
		device_printf(sc->sc_dev, "Error malloc'ing status page\n");
		return ENOMEM;
	}

	sc->sc_video_cdev = make_dev(&saa_video_cdevsw, 
					unit,
					UID_ROOT, GID_WHEEL, 0600,
					"saa%d", unit);

	REG_SET(REG_MCR, (MCR_VPLLE | MCR_EXOSC | MCR_EVFE1 | MCR_EVFE2));
	return 0;
}


int
saa_video_dnit(struct saa_softc *sc)
{
	if (sc->sc_sts) {
		free(sc->sc_sts, M_SAA_VCAP);
		sc->sc_sts = 0;
	}

	if (sc->sc_video_cdev)
		destroy_dev(sc->sc_video_cdev);

	return 0;
}


static int 
saa_videodev_open(DPTYPE dev, int oflags, int dtype, struct thread *td)
{
	struct saa_softc *sc;
	int unit;

	unit = minor(dev);
	sc = devclass_get_softc(saa_devclass, unit);
	//device_busy(devclass_get_device(saa_devclass, unit));
	return 0;
}

static int 
saa_videodev_close(DPTYPE dev, int fflags, int dtype, struct thread *td)
{
	struct saa_softc *sc;
	int unit, t;

	unit = minor(dev);
	sc = devclass_get_softc(saa_devclass, unit);
	if (sc->sc_video_flags & SCVF_OVERLAY)
		saa_video_overlay_off(sc);
	else if(sc->sc_video_flags & SCVF_CAPTURE)
		saa_video_capture_off(sc);

	for (t = 0; t < 2; ++t)
		if (sc->sc_video_capbuf_allocflag[t]) {
			if (saa_video_capbuf_dnit(sc, t) != 0)
				device_printf(sc->sc_dev,
						"device_close: error releasing"
						" capture buffer. Mem. leak\n");
			else
				sc->sc_video_capbuf_allocflag[t] = 0;	
		}
	//device_unbusy(devclass_get_device(saa_devclass, unit));
	return 0;
}

static int 
saa_videodev_ioctl(DPTYPE dev, u_long cmd,  caddr_t data, int fflags, 
		struct thread *td)
{
	struct saa_softc *sc;
	int unit, rval = 0;

	unit = minor(dev);
	sc = devclass_get_softc(saa_devclass, unit);

	switch(cmd) {
	case GETVIDEOSTD:
	{
		struct saa_video_std *s = (struct saa_video_std*) data;
		*s = sc->sc_video_std;
		break;
	}
	case SETVIDEOSTD:
	{
		struct saa_video_std *s = (struct saa_video_std*) data;
		sc->sc_video_std = *s;
		saa_video_setstd(sc, &sc->sc_video_std, 0);
		saa_video_setstd(sc, &sc->sc_video_std, 1);
		if (sc->sc_video_flags & SCVF_ON) {
			saa_video_setup_scaler(sc, 0);
			saa_video_setup_scaler(sc, 1);
		}
		break;
	}
	case GETVIDEOOPT:
	{
		struct saa_video_opt *o = (struct saa_video_opt*) data;
		*o = sc->sc_video_opt;
		break;
	}
	case SETVIDEOOPT:
	{
		struct saa_video_opt *o = (struct saa_video_opt*) data;
		sc->sc_video_opt = *o;
		saa_video_setopt(sc, &sc->sc_video_opt);
		if (sc->sc_video_flags & SCVF_ON) {
			saa_video_setup_scaler(sc, 0);
			saa_video_setup_scaler(sc, 1);
		}
		break;
	}
	case OVERLAYON:
	{
		struct saa_video_spec *o = (struct saa_video_spec*) data;
		sc->sc_video_spec = *o;
		if (sc->sc_video_flags & SCVF_ON) {
			rval = EBUSY;	
			break;
		}
		/* XXX Only status signal is valid in overlay mode */
		if (o->svs_flags & VSPEC_FLAG_STATUS_SIGNAL) {
			sc->sc_video_capbuf_sigpid = td->td_proc->p_pid;
		} else
			sc->sc_video_capbuf_sigpid = 0;
		rval = saa_video_overlay_on(sc, &sc->sc_video_spec);
		break;
	}
	case OVERLAYOFF:
		if (!(sc->sc_video_flags & SCVF_OVERLAY))
			break;
		saa_video_overlay_off(sc);
		break;	

	case CAPTUREON:
	{
		struct saa_video_spec *o = (struct saa_video_spec*) data;
		if (sc->sc_video_flags & SCVF_ON) {
			rval = EBUSY;
			break;
		}
		sc->sc_video_spec = *o;
		if (o->svs_flags & (VSPEC_FLAG_CAPTURE_SIGNAL | 
				  	VSPEC_FLAG_STATUS_SIGNAL))
			sc->sc_video_capbuf_sigpid = td->td_proc->p_pid;
		else
			sc->sc_video_capbuf_sigpid = 0;
		rval = saa_video_capture_on(sc, &sc->sc_video_spec, td);
		break;
	}
	case CAPTUREOFF:
		if (!(sc->sc_video_flags & SCVF_CAPTURE))
			break;
		saa_video_capture_off(sc);
		sc->sc_video_capbuf_sigpid = 0;
		break;
	case SETCLIPINFO:
	{
		struct saa_video_clipinfo *o = (struct saa_video_clipinfo*)data;
		if ((sc->sc_video_flags & SCVF_ON) == 0)
			rval = EDOOFUS;
		else {
			saa_load_cliplist(sc, 
					&o->svci_clipmode,
					&o->svci_cliplist);
		}
		break;
	}
	case VIDEODMAPAUSE:
		if ((sc->sc_video_flags & SCVF_ON) == 0) {
			rval = EDOOFUS;
		} else {
			if ((sc->sc_video_flags & SCVF_DMA_PAUSED) == 0) {
				sc->sc_video_dmasave = REG_READ(REG_MCR);
				REG_CLR(REG_MCR, MCR_CHAN0_TE | MCR_CHAN1_TE |
						MCR_CHAN4_TE | MCR_CHAN5_TE);
				sc->sc_video_flags |= SCVF_DMA_PAUSED;
			}
		}
		break;
	case VIDEODMACONT:
		if ((sc->sc_video_flags & SCVF_ON) == 0) {
			rval = EDOOFUS;
		} else {
			if (sc->sc_video_flags & SCVF_DMA_PAUSED) {
				/* XXX Note use of set */
				sc->sc_video_dmasave &= MCR_CHAN0_TE |
							MCR_CHAN1_TE |
							MCR_CHAN4_TE |
							MCR_CHAN5_TE;
				REG_SET(REG_MCR, sc->sc_video_dmasave);
				sc->sc_video_flags &= ~SCVF_DMA_PAUSED;
			}
		}
		break;
	case GETVIDEOSTS:
	{
		uint32_t *o = (uint32_t*)data;
		*o = REG_READ(BREG_VIDEO_STS1) & 0xffff;
		break;
	}
	case GETVIDEOVPTR:
	{
		uint32_t *o = (uint32_t*)data;
		*o = REG_READ(REG_SLICE0_STS);
		break;
	}
	case GETGPIODIR:
		*(uint32_t*)data = REG_READ(REG_GPIO_DIR);
		break;
	case SETGPIODIR:
		REG_WRITE(REG_GPIO_DIR, *(uint32_t*)data);
		break;
	case GETGPIODATA:
		*(uint32_t*)data = REG_READ(REG_GPIO_DATA);
		break;
	case SETGPIODATA:
		REG_WRITE(REG_GPIO_DATA, *(uint32_t*)data);
		break;
	default:
		rval = EINVAL;
		break;
	}
	return rval;
}


static void
saa_video_overlay_off(struct saa_softc *sc)
{
	REG_CLR(REG_IEN, sc->sc_video_spec.svs_cap_sigmask);
	REG_CLRB(BREG_REGION_EN, REGION_EN_VIDEO_A);
	REG_CLR(REG_MCR, MCR_CHAN0_TE);
	sc->sc_video_flags &= ~(SCVF_ON | SCVF_OVERLAY | SCVF_DMA_PAUSED);
}


/*
 * Overlay directly to a memory location via DMA
 * DANGEROUS FUNCTION as it can overwirte any physical
 * memory location. Once setup overlay uses 0% CPU 
 * it requires only the PCI bus
 * Works only with packed modes, YUV packed modes are
 * supported
 */

static int
saa_video_overlay_on(struct saa_softc *sc, struct saa_video_spec *o)
{
	uint8_t b;
	int t = SAA_CHAN_TASKA;

	/* Overlay supported in Non Planar modes only */
	if (o->svs_vfmt & VFMT_PLANAR)
		return EINVAL;
	
	if ((o->svs_flags & VSPEC_FLAG_OVERLAY_DROPODD) != 0 &&
		(o->svs_flags & VSPEC_FLAG_OVERLAY_DROPEVEN) != 0)
		return EINVAL;

	saa_video_setopt(sc, &sc->sc_video_opt);
	saa_video_setstd(sc, &sc->sc_video_std, t);

	b = TASK_COND_UEVEN | TASK_COND_UODD;
	if (o->svs_flags & VSPEC_FLAG_OVERLAY_DROPODD)
		b = TASK_COND_UEVEN;
	else if (o->svs_flags & VSPEC_FLAG_OVERLAY_DROPEVEN)
		b = TASK_COND_UODD;

	REG_WRITEB(BREG_TASK_COND(t), 
			(o->svs_flags & VSPEC_FLAG_FRAMELOCK ? 
							TASK_COND_SLCK : 0) |
			TASK_COND_U50 |  TASK_COND_U60 | b);
	if (b == (TASK_COND_UEVEN | TASK_COND_UODD))
		b = 2;
	else
		b = 1;
	REG_WRITEB(BREG_TASK_FEILD(t),
		((o->svs_cap_frameskip[t] << TASK_FEILD_FSKIP_SHIFT) & 
						TASK_FEILD_FSKIP_MASK) |  
		((b <<  TASK_FEILD_COUNT_SHIFT) & TASK_FEILD_COUNT_MASK));

	/* CSCON - RGB to YUV filter */
	/* We use CLRB, SETB, to avoid overwriting VBIRAWSRC */
	if ((o->svs_vfmt & VFMT_YUV) == 0) {
		REG_CLRB(BREG_TASK_DPATH(t), TASK_DPATH_CSCON);
		REG_SETB(BREG_TASK_DPATH(t), TASK_DPATH_MACON);
	} else {
		/* 
		 * instead of CLRB(...) we could use 
		 * SETB(TASK_DPATH_MACON | TASK_DPATH_CSCON)
		 * which is a must when we use GACON (XXX GACON not impl.)
		 */
			REG_CLRB(BREG_TASK_DPATH(t), 
					(TASK_DPATH_CSCON | TASK_DPATH_MACON));
	}

	REG_WRITEW(WREG_TASK_VIDEO_XOUTLEN(t), o->svs_buf_width);
	REG_WRITEW(WREG_TASK_VIDEO_YOUTLEN(t), o->svs_buf_height);

	/* XXX Will disable clipping, ie. will set CLIPMODE to OFF */
	REG_WRITEB(BREG_TASK_VIDEO_FMT(t), 
					(o->svs_vfmt & VFMT_FMT_MASK) |
					((o->svs_flags & VSPEC_FLAG_DITHER) 
								? 0x80 : 0));
	/* XXX Will reset INTLCD and CLINV */
	REG_WRITEB(BREG_VIDEO_UVL, 
			((o->svs_vfmt & VFMT_UVP_MASK) >> VFMT_UVP_SHIFT) &
			VIDEO_UVL_MASK);
				
	saa_video_setup_scaler(sc, t);

	saa_load_cliplist(sc, &o->svs_clipmode, &o->svs_cliplist);

	REG_WRITE(REG_DMA_BASE0(t), 
			(int)o->svs_buf_base + o->svs_buf_offset[0][0]);
	REG_WRITE(REG_DMA_BASE1(t), 
			(int)o->svs_buf_base + o->svs_buf_offset[1][0]);
	REG_WRITE(REG_DMA_PITCH(t), o->svs_buf_pitch[0]);
	REG_WRITE(REG_DMA_CTRL(t), 
			((0x7 << DMA_CTRL_BURST_SHIFT) & DMA_CTRL_BURST_MASK) |
			o->svs_cap_swapbytes);

	FIFO_SET_SIZE(896, 0, 0, 128);
	REG_WRITEB(BREG_REGION_EN, REGION_EN_VIDEO_A);
	REG_SET(REG_MCR, MCR_CHAN0_TE);

	if (o->svs_flags & VSPEC_FLAG_STATUS_SIGNAL) {
		o->svs_cap_sigmask &= STS_ALL;
		REG_SET(REG_IEN, o->svs_cap_sigmask);
	} else {
		o->svs_cap_sigmask = 0;
		REG_CLR(REG_IEN, STS_ALL);
	}

	sc->sc_video_flags |= SCVF_ON | SCVF_OVERLAY;
	return 0;
}


/*
 * Setup horizontal and vertical video scaling, influenced
 * by VSPEC_CAPTURE_NO_HSCALE, VSPEC_CAPTURE_NO_VSCALE
 * settings in sc->sc_video_spec.svs_vflags
 * Can be called while capture is on
 * XXX Overlay mode + scaling untested, may cause severe
 *	memory corruption
 */
 
static void
saa_video_setup_scaler(struct saa_softc *sc, int t)
{
	uint16_t maxw, maxh;
	/* XXX '* 2' works only for interlaced output */
	maxh = (sc->sc_video_std.std_video_yinstop - 
			sc->sc_video_std.std_video_yinstart + 1) * 2;
	maxw = sc->sc_video_std.std_video_xinstop -
			sc->sc_video_std.std_video_xinstart + 1;

	/* XXX Horizontal Pre-scale not supported */
	REG_WRITEB(BREG_TASK_XPSC(t), 1);
	REG_WRITEB(BREG_TASK_XPSC_XACL(t), 0);
	REG_WRITEB(BREG_TASK_XPSC_CTRL(t), 0);
	REG_WRITEB(BREG_TASK_FIRFLTR_CTRL(t), 0);

	if (sc->sc_video_spec.svs_flags & VSPEC_FLAG_NO_HSCALE)
		REG_WRITEW(WREG_TASK_VIDEO_XSC(t), 1024);
	else
		REG_WRITEW(WREG_TASK_VIDEO_XSC(t),
					(maxw * 1024) / 
					sc->sc_video_spec.svs_buf_width);

	if (sc->sc_video_spec.svs_flags & VSPEC_FLAG_NO_VSCALE) {
		REG_WRITEW(WREG_TASK_YSC(t), 1024);
		REG_WRITEB(BREG_TASK_YFLTR_CNTRL(t), 0x00);
	} else {
		uint16_t yscale;
		yscale = (maxh * 1024) / sc->sc_video_spec.svs_buf_height;
		REG_WRITEW(WREG_TASK_YSC(t), yscale);
		if (yscale < 2048)
			REG_WRITEB(BREG_TASK_YFLTR_CNTRL(t), 0x00);
		else
			REG_WRITEB(BREG_TASK_YFLTR_CNTRL(t), 
					TASK_YFLTR_CNTRL_YMODE);
	}
}


static int 
saa_videodev_mmap(DPTYPE dev, vm_offset_t offset, vm_paddr_t *paddr, int nprot)
{
	struct saa_softc *sc;
	int unit;
	int f;
	uint32_t *pta;

	unit = minor(dev);
	sc = devclass_get_softc(saa_devclass, unit);

	if (offset == SAA_MMAP_STS_OFFSET) {
		/* Special case, video status page map */
		/* sts is malloced on a page boundary (in video_init) */

		*paddr = vtophys(sc->sc_sts);
		return 0;
	}

	if ((sc->sc_video_flags & SCVF_CAPTURE) == 0 ||
		(sc->sc_video_spec.svs_flags & VSPEC_FLAG_CAPTURE_SHM) != 0)
		return EDOOFUS;

	if (nprot & (PROT_EXEC | PROT_WRITE))
		return EPERM;

	if ((offset >= SAA_MMAP_T0_OFFSET) &&
	    (offset < (SAA_MMAP_T0_OFFSET + SAA_MMAP_T0_SZ))) {
		f = 0;				/* Frame 1 */
		offset -= SAA_MMAP_T0_OFFSET;
		offset >>= 12;
	} else if ((offset >= SAA_MMAP_T1_OFFSET) &&
	    	   (offset < (SAA_MMAP_T1_OFFSET + SAA_MMAP_T1_SZ))) {
		f = 1;				/* Frame 2 */
		offset -= SAA_MMAP_T1_OFFSET;
		offset >>= 12;
	} else
		return EINVAL;

	
	if (sc->sc_video_capbuf_allocflag[f] == 0)
		return EDOOFUS;

	/* 
	 * Make sure offset is within our capture buffer
	 * It is invariable that both buffers are of the
	 * same size (incase of packed TASKA+B capture)
	 * hence we dont't need svs_cap_bufsz[t]
	 */
	if (offset > (sc->sc_video_spec.svs_cap_bufsz - 4096)) {
		device_printf(sc->sc_dev, "Invalid mmap offset %d, max: %d\n", 
				offset,
				(sc->sc_video_spec.svs_cap_bufsz - 4096));
		return EINVAL;
	}

	pta = &(sc->sc_pta[f * 1024]);
	*paddr = pta[offset];
	return 0;
}

/*
 * Switch between buffers in MULTIBUF mode
 * (see note above capture_on)
 */

void
saa_video_multibuf_switchbuf(struct saa_softc *sc)
{
	int t;
	int offset;
	struct saa_video_spec *o;

	o = &sc->sc_video_spec;
	t = sc->sc_sts->ssts_video_lasttask;
	
	/* Note increment */
	if (++sc->sc_sts->ssts_video_mb_cur[t] >=
				sc->sc_sts->ssts_video_mb_count)
		sc->sc_sts->ssts_video_mb_cur[t] = 0;

	offset = sc->sc_sts->ssts_video_mbufsz *
				sc->sc_sts->ssts_video_mb_cur[t];

	if (sc->sc_video_spec.svs_vfmt & VFMT_PLANAR) {
		REG_WRITE(REG_DMA_BASE0(SAA_CHAN_YUV_Y),
			(int)o->svs_buf_base + o->svs_buf_offset[0][0] + 
			offset);
		REG_WRITE(REG_DMA_BASE1(SAA_CHAN_YUV_Y),
			(int)o->svs_buf_base + o->svs_buf_offset[1][0] +
			offset);

		REG_WRITE(REG_DMA_BASE0(SAA_CHAN_YUV_U), 
			(int)o->svs_buf_base + o->svs_buf_offset[0][1] +
			offset);
		REG_WRITE(REG_DMA_BASE1(SAA_CHAN_YUV_U), 
			(int)o->svs_buf_base + o->svs_buf_offset[1][1] +
			offset);

		REG_WRITE(REG_DMA_BASE0(SAA_CHAN_YUV_V), 
			(int)o->svs_buf_base + o->svs_buf_offset[0][2] +
			offset);
		REG_WRITE(REG_DMA_BASE1(SAA_CHAN_YUV_V), 
			(int)o->svs_buf_base + o->svs_buf_offset[1][2] +
			offset);
	} else {
		REG_WRITE(REG_DMA_BASE0(t), 
			(int)o->svs_buf_base + o->svs_buf_offset[0][0] + 
				offset);
		REG_WRITE(REG_DMA_BASE1(t),
			(int)o->svs_buf_base + o->svs_buf_offset[1][0] + 
				offset);
	}
}


static void
saa_load_cliplist(struct saa_softc *sc, struct saa_video_clipmode *svcm,
			struct saa_video_cliplist *svcl)
{
	uint32_t i, j, xcount, ycount, mcr_save;
	uint16_t v, w;
	struct entry {
		uint16_t	wi;
		uint16_t	vi;
	} xe[SAA_MAX_CLIPS * 2], ye[SAA_MAX_CLIPS * 2], temp;

	/* SETB must, as not to wipe out packing modes and dither info
	 * set using BREG_TASK_VIDEO_FMT(x)
	 */
	REG_SETB(BREG_TASK_VIDEO_CLIPMODE(0), svcm->svcm_mode);
	REG_SETB(BREG_TASK_VIDEO_CLIPMODE(1), svcm->svcm_mode);
	i = svcm->svcm_flag_intlcd ? VIDEO_UVL_INTLCD : 0;
	i |= svcm->svcm_flag_clinv ? VIDEO_UVL_CLINV : 0;
	REG_SETB(BREG_VIDEO_UVL, i);
	switch (svcm->svcm_mode) {
	case CLIPMODE_ALPHA:
		REG_WRITEB(BREG_CLIP_ACL, svcm->svcm_info.ai.acl);
		REG_WRITEB(BREG_CLIP_ANCL, svcm->svcm_info.ai.ancl);
		break;
	case CLIPMODE_FIXED:
		REG_WRITEB(BREG_CLIP_CLCR, svcm->svcm_info.fi.red);
		REG_WRITEB(BREG_CLIP_CLCG, svcm->svcm_info.fi.green);
		REG_WRITEB(BREG_CLIP_CLCB, svcm->svcm_info.fi.blue);
	default:
		break;
	}

	if (svcm->svcm_mode == CLIPMODE_OFF || svcl->svcl_count == 0) {
		/* No need to stop DMA */
		for (i = 0; i < (SAA_MAX_CLIPS * 2); ++i) {
			REG_WRITE(REG_CLIP_X(i), 0);
			REG_WRITE(REG_CLIP_Y(i), 0);
		}
	} else {
		svcl->svcl_count = svcl->svcl_count > SAA_MAX_CLIPS ?
							SAA_MAX_CLIPS :
							svcl->svcl_count;
		for (i = 0; i < svcl->svcl_count; ++i) {
			w = (1 << (SAA_MAX_CLIPS - i - 1));
			/* 1 << 15, signifies win start 
			 * 0 << 15, signifies win end
			 */
			xe[i * 2].wi = w | (1 << 15);
			xe[i * 2].vi = svcl->svcl_clip[i].x;

			ye[i * 2].wi = w | (1 << 15);
			ye[i * 2].vi = svcl->svcl_clip[i].y;

			xe[i * 2 + 1].wi = w;
			xe[i * 2 + 1].vi = svcl->svcl_clip[i].x + 
							svcl->svcl_clip[i].w;

			ye[i * 2 + 1].wi = w;
			ye[i * 2 + 1].vi = svcl->svcl_clip[i].y + 
							svcl->svcl_clip[i].h;
		}
		/* Bubble Sort xe and ye */
		ycount = svcl->svcl_count * 2 - 1;
		xcount = svcl->svcl_count * 2;
		for (i = 0; i < ycount; ++i) {
			for (j = i + 1; j < xcount; ++j) {
				if (xe[i].vi > xe[j].vi) {
					temp = xe[i];
					xe[i] = xe[j];
					xe[j] = temp;
				}
				if (ye[i].vi > ye[j].vi) {
					temp = ye[i];
					ye[i] = ye[j];
					ye[j] = temp;
				}
			}
		}

		/* Stop DMA prior to writing new cliplist */
		mcr_save = REG_READ(REG_MCR);
		REG_CLR(REG_MCR, MCR_CHAN0_TE | MCR_CHAN1_TE |
					MCR_CHAN4_TE | MCR_CHAN5_TE);

		/* Write while merging, xe */
		xcount = svcl->svcl_count * 2;
		ycount = 0;
		w = 0;
		for (i = 0; i < xcount;) {
			v = xe[i].vi;
			do {
				if (xe[i].wi & (1 << 15)) { 
					/* start of win, switch on bit */
					w |= (~(1 << 15)) & xe[i].wi;
				} else {
					/* end of win, switch off bit */
					w &= ~xe[i].wi;
				}
				++i;
			} while(xe[i].vi == v);
			REG_WRITE(REG_CLIP_X(ycount), ((uint32_t)v << 16 | w));
			++ycount;
		}
		for (;ycount < (SAA_MAX_CLIPS * 2); ++ycount)
			REG_WRITE(REG_CLIP_X(ycount), 0);

		/* Write while merging, ye */
		xcount = svcl->svcl_count * 2;
		ycount = 0;
		w = 0;
		for (i = 0; i < xcount;) {
			v = ye[i].vi;
			do {
				if (ye[i].wi & (1 << 15)) { 
					/* start of win, switch on bit */
					w |= (~(1 << 15)) & ye[i].wi;
				} else {
					/* end of win, switch off bit */
					w &= ~ye[i].wi;
				}
				++i;
			} while(ye[i].vi == v);
			REG_WRITE(REG_CLIP_Y(ycount), ((uint32_t)v << 16 | w));
			++ycount;
		}
		for (;ycount < (SAA_MAX_CLIPS * 2); ++ycount)
			REG_WRITE(REG_CLIP_Y(ycount), 0);

		/* Start DMA */
		REG_WRITE(REG_MCR, mcr_save);
	}
}
