#include <sys/param.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>

#include "saa713x_ioctl.h"

#define DEFAULT_SAU_DEVPATH	"/dev/sau0"
#define DEFAULT_DSP_DEVPATH	"/dev/dsp0"

#define FLAG_OUTFILE		(1 << 0)

struct GLOBAL {
	struct saa_status 	*sts;
	uint8_t 		*abuf;
	int 			devfd;
	int 			sndfd;
	uint32_t 		ctr_frame;
	uint32_t		flags;
	char			*devpath;
	char			*outpath;
} G;


static struct saa_audio_opt audio_opt = { /* PAL A2 FM Defaults */
	.sao_flag_idarea = 0,
	.sao_idmod = 0,

	.sao_ch1mode = CH1MODE_FM,
	.sao_ch2mode = CH2MODE_FM,
	.sao_fltbw = FLTBW_C1MIN_C2MIN,

	.sao_siffreq1 = 0x729555,
	.sao_siffreq2 = 0x77A101,

	.sao_flag_hpen = 0,
	.sao_sifp = 0,
	.sao_sifa = SIFA_N10DB,

	.sao_sifgain = 0,
	.sao_flag_sifagc_off = 0,
	.sao_flag_sifagc_slow = 0,

	.sao_flag_fmdsch1 = 0,
	.sao_fmdsch1 = 0,
	.sao_flag_fmdsch2 = 0,
	.sao_fmdsch2 = 0,

	.sao_flag_idswfm = 1,		/* Auto DMSel ON */
	.sao_fmdmsel = 0,

	.sao_dac_sdos = SDOS_FMAM_FMAM,
	.sao_dac_avl = AVL_OFF,
	.sao_dac_ogain = SDGS_0DB,

	.sao_flag_aasdma = 1,
	.sao_dsp_csm = CSM_LEFT_RIGHT,

	.sao_flag_ch1_2vrms = 0,
	.sao_flag_ch2_2vrms = 0
};


static struct saa_audio_spec audio_spec = {
	.sas_audsel = AUDSEL_SIF,
	.sas_sfs = SFS_SIF_32KHZ,
	.sas_icsch = 0,
	.sas_cap_sampcount = 512,
	.sas_audfmt = AUDFMT_16BIT,
	.sas_audsamp = AUDSAMP_STEREO,
	.sas_flag_samp2complement = 1,
	.sas_flags = ASPEC_FLAG_CAPTURE_SIGNAL,
	.sas_cap_mb_limit = 0,
	.sas_aclkfreq = ACLKFREQ_32p11MHZ,
};


static void
audio_notify_multibuf(int sig)
{
	int offset;

	while (G.ctr_frame < G.sts->ssts_audio_framectr) {
                offset = G.ctr_frame % G.sts->ssts_audio_mb_count;
                offset *= G.sts->ssts_audio_framesz;
                write(G.sndfd, 
                        G.abuf + offset, G.sts->ssts_audio_framesz);
                ++G.ctr_frame;
	}
}


static void
audio_notify(int sig)
{
	write(G.sndfd, 
		G.abuf +(G.sts->ssts_audio_framesz * G.sts->ssts_audio_lastbuf),
		G.sts->ssts_audio_framesz);
}


static const char *opts = "i:s:vNMD:A:o:c:";

static const char *usage = 
	"**Sampling rate is 32KHz and Sample format is 16 bit signed\n\n"
    	"-i <input>		: Audio Source:\n"
	"			    1 - Audio Input 1\n"
	"			    2 - Audio Input 2\n"
	"			    s - SIF (Tuner) Audio (default)\n"
	"			    m - SIF (Tuner) Mono (Left) Audio\n"
	"			    n - SIF (Tuner) Mono (Right) Audio\n"
	"-s <N>			: Caputre buffer size, N is:\n"
	"			    128 <= N <= (512 * 1024) (default: 512)\n"
	"-v 			: Lock audio to video: (default: false)\n"
	"-N 			: Video standard is NTSC (default: PAL)\n"
	"-M 			: Enable Multibuf mode (default: false)\n"
	"-D <path>		: SAU device path (default: /dev/sau0)\n"
	"-A <path>		: Audio device path (default: /dev/dsp0)\n"
	"-o <path>		: Output to file\n"
	"-c <1|2>		: Audio clock frequency (default 32.11MHz)\n"
	"			    1 - 32.110 MHz\n"
	"			    2 - 24.576 MHz\n";


int
main(int argc, char **argv)
{
	struct sigaction sa;
	sigset_t ss;
	int i, ch;

	while ( (ch = getopt(argc, argv, opts)) != -1) {
		switch (ch) {
		case 'i':
			switch(optarg[0]) {
			case '1':
				audio_spec.sas_audsel = AUDSEL_BBAND;
				audio_spec.sas_sfs = SFS_BBAND_32KHZ;
				audio_spec.sas_icsch = 0;
				break;
			case '2':
				audio_spec.sas_audsel = AUDSEL_BBAND;
				audio_spec.sas_sfs = SFS_BBAND_32KHZ;
				audio_spec.sas_icsch = 1;
				break;
			case 's':
				audio_spec.sas_audsel = AUDSEL_SIF;
				audio_spec.sas_sfs = SFS_SIF_32KHZ;
				break;
			case 'm':
				audio_spec.sas_audsel = AUDSEL_SIF;
				audio_spec.sas_sfs = SFS_SIF_32KHZ;
				audio_spec.sas_audsamp = AUDSAMP_MONO_LEFT;
				break;
			case 'n':
				audio_spec.sas_audsel = AUDSEL_SIF;
				audio_spec.sas_sfs = SFS_SIF_32KHZ;
				audio_spec.sas_audsamp = AUDSAMP_MONO_RIGHT;
				break;
			default:
				errx(EINVAL, 
					"Invalid input source '%s'\n\n%s", 
					optarg, usage);
			}
			break;
		case 's':
			audio_spec.sas_cap_sampcount = atoi(optarg);
			if (audio_spec.sas_cap_sampcount < 128 ||
				audio_spec.sas_cap_sampcount > (512 * 1024))
				errx(EINVAL, "Invalid sample count %s\n\n%s",	
					optarg, usage);
			break;
		case 'v':
			audio_spec.sas_flags |= ASPEC_FLAG_VLOCK;
			break;
		case 'N':
			audio_spec.sas_flags |= ASPEC_FLAG_NTSC;
			break;
		case 'M':
			audio_spec.sas_flags |= ASPEC_FLAG_CAPTURE_MULTIBUF;
			break;
		case 'D':
			G.devpath = optarg;
			break;
		case 'A':
			G.outpath = optarg;
			G.flags &= (~FLAG_OUTFILE);
			break;
		case 'o':
			G.outpath = optarg;
			G.flags |= FLAG_OUTFILE;
			break;
		case 'c':
			switch (optarg[0]) {
				case '1':
					audio_spec.sas_aclkfreq = 
						ACLKFREQ_32p11MHZ;
					break;
				case '2':
					audio_spec.sas_aclkfreq = 
						ACLKFREQ_24p576MHZ;
					break;
				default:
					errx(EINVAL, "Invalid audio clock "
					 "frequency: %s\n\n%s", 
					 optarg, usage);
			}
			break;
		default:
			errx(EINVAL, "Invalid option %c\n\n%s",
				ch, usage);
		}
	}

	if (G.devpath == 0)
		G.devpath = DEFAULT_SAU_DEVPATH;
	
	if ( (G.devfd = open(G.devpath, O_RDWR)) == -1)
		err(errno, "Error opening audio device %s", G.devpath);

	if (G.flags & FLAG_OUTFILE) {
		if (G.outpath == 0)
			G.sndfd = 1;	// standard output
		else {
			if ( (G.sndfd = open(G.outpath, 
						O_WRONLY|O_CREAT,
						0644)) == -1)
				err(errno, "Error opening output file %s", 
					G.outpath);
		}
	} else {
		if (G.outpath == 0)
			G.outpath = DEFAULT_DSP_DEVPATH;

		if ( (G.sndfd = open(G.outpath, O_RDWR)) == -1)
				err(errno, "Error opening DSP device %s", 
					G.outpath);
		i = 32000;
        	if (ioctl(G.sndfd, SNDCTL_DSP_SPEED, &i) == -1)
                	err(errno, "Error setting DSP speed to 32KHz");

		i = AFMT_S16_LE;
		if (audio_spec.sas_audsamp == AUDSAMP_STEREO)	
			i |= AFMT_STEREO;

        	if (ioctl(G.sndfd, SNDCTL_DSP_SETFMT, &i) == -1)
                	err(errno, "Error setting DSP format");
	}


	if (ioctl(G.devfd, SETAUDIOOPT, &audio_opt) == -1)
		err(errno, "SAU device SETAUDIOOPT failure");

	if ( (G.sts = mmap(0, SAA_MMAP_STS_SZ, PROT_READ, MAP_NOCORE,
                                G.devfd,
                                SAA_MMAP_STS_OFFSET)) == MAP_FAILED)
                err(errno, "Error memory mapping SAU device status buffer");

	if ( (G.abuf = mmap(0, 1048576, PROT_READ, 0,
				G.devfd, SAA_MMAP_ACAP_OFFSET)) == MAP_FAILED)
		err(errno, "Error memory mapping SAU device frame");

	if (audio_spec.sas_flags & ASPEC_FLAG_CAPTURE_MULTIBUF)
        	sa.sa_handler = audio_notify_multibuf;
	else
        	sa.sa_handler = audio_notify;

        sa.sa_flags = 0;
        sigemptyset(&ss);
        sa.sa_mask = ss;

        if (sigaction(SIGUSR2, &sa, NULL) != 0)
                err(errno, "Error setting up signal handler for SIGUSR2");

	if (ioctl(G.devfd, ACAPTUREON, &audio_spec) == -1)
		err(errno, "SAU device ACAPTUREON failure");

	while (1)
		pause();

	return 0;
}
