#include <sys/param.h>
#include <sys/mman.h>
#include <sys/shm.h>
#include <fcntl.h>
#include <err.h>
#include <errno.h>
#include <unistd.h>
#include <sysexits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>

#include "tvv.h"
#include "tvv_clip.h"
#include "tvv_util.h"

/* Sound option DAC -> SndCard via /dev/pcm */

const char *optstr = "Xf:D:V:F:a:A:M";
const char *usage = 
	"-X			: Enable DSHM\n"
	"-M			: Enable MULTIBUF\n"
	"-f <device>		: Capture device, /dev/saa0 (default)\n"
    	"-D <mode>		: Playback using DGA, mode is:\n"
	"			    w - Weave\n" 
	"			    b - Bob\n"
	"			    o - Drop field 1 (odd lines)\n"
	"			    e - Drop field 2 (even lines)\n"
    	"-V <mode>		: Playback using Xv (default), mode is:\n"
	"			    w - Weave (default)\n" 
	"			    r - Weave-RT\n" 
	"			    b - Bob\n"
	"			    o - Drop field 2 (odd lines)\n"
	"			    e - Drop field 1 (even lines)\n"
	"-F <format>		: Xv Playback format, format is:\n"
	"			    u - YUY2\n"
	"			    y - YV12\n"
	"			    r - RGB\n"
	"-a <mode>		: Audio Lineout, output selection:\n"
	"			    d - DAC\n"
	"			    1 - Linein 1, stereo\n"
	"			    2 - Linein 2, stereo (default)\n"
	"			    3 - Linein 1, left mono\n"
	"			    4 - Linein 1, right mono\n"
	"			    5 - Linein 2, left mono\n"
	"			    6 - Linein 2, right mono\n"
	"-A <device>		: Audio device,  /dev/sau0 (default)\n";

struct privdata {
	struct vinfo *vi;
	struct ainfo *ai;
	struct xinfo *xi;
};


static void	alloc_defaults(struct vinfo **vi, 
			struct ainfo **ai, struct xinfo **xi);

static void 	parse_args(int argc, char **argv, 
			struct vinfo *vi, struct ainfo *ai, struct xinfo *xi);

static __inline void 
		xv_setmodeinfo(uint32_t fourcc, 
			struct vinfo *vi, struct xinfo *xi);

static void 	audio_init(struct ainfo *ai);
static void 	dga_init(struct vinfo *vi, struct xinfo *xi);
static void 	xv_init(struct vinfo *vi, struct xinfo *xi);

static void 	xv_status_notify(int sig);
static void 	xv_frame_notify(int sig);
static void 	xv_multibuf_notify(int sig);
static void 	terminate_notify(int sig);

static void	start_video(struct vinfo *vi, struct xinfo *xi);
static void 	stop_video(struct vinfo *vi, struct xinfo *ai);
static void 	reset_viewport(struct vinfo *vi, struct xinfo *xi);

static void 	event_handler(Widget w, XtPointer priv, 
			XEvent *e, Boolean *cont);

static void	get_source_dimensions(struct vinfo *vi, 
			int *width, int *height);

/* 
 * Global variables required due to processing done in signal handler
 */
struct vinfo *gvi;
struct ainfo *gai;
struct xinfo *gxi;

int
main(int argc, char **argv)
{
	XSetWindowAttributes cattr;
	XWindowAttributes wattr;
	XEvent e;
	XErrorHandler eh;
	struct privdata p;
	Screen *scr;
	int i;

	alloc_defaults(&p.vi, &p.ai, &p.xi);
	gvi = p.vi; gai = p.ai; gxi = p.xi;

	gxi->xi_win = XtVaOpenApplication(&gxi->xi_app, "TVV",
					NULL, 0,
					&argc, argv,
					NULL,
					applicationShellWidgetClass,
					NULL);
	scr = DefaultScreenOfDisplay(XtDisplay(gxi->xi_win));
        gxi->xi_scr_width = WidthOfScreen(scr);
        gxi->xi_scr_height = HeightOfScreen(scr);

	parse_args(argc, argv, gvi, gai, gxi);

	if ( (gvi->vi_dfd = open(gvi->vi_dname, O_RDWR)) == -1)
		err(errno, "error opening %s", gvi->vi_dname);

	get_source_dimensions(gvi, &gvi->vi_src_width, &gvi->vi_src_height);
	gxi->xi_win_width = gvi->vi_src_width;
	gxi->xi_win_height = gvi->vi_src_height;

	audio_init(gai);

	if (gxi->xi_dgainfo->dga_mode) {
		/* SAA Clipping Parameters */
		gvi->vi_svs.svs_clipmode.svcm_mode = CLIPMODE_RECT;
		gvi->vi_svs.svs_clipmode.svcm_flag_clinv = 0;
		gvi->vi_svs.svs_clipmode.svcm_flag_intlcd = 1;
		if (gxi->xi_dgainfo->dga_mode == 'o' || 
		    gxi->xi_dgainfo->dga_mode == 'e')
			gvi->vi_svs.svs_clipmode.svcm_flag_intlcd = 0;
		dga_init(gvi, gxi);
	}

	XtAddEventHandler(gxi->xi_win,
				StructureNotifyMask | VisibilityChangeMask |
				 ExposureMask,
				FALSE,
				event_handler, 
				&p);

	XtVaSetValues(gxi->xi_win, 
			XtNwidth, gxi->xi_win_width,
			XtNheight, gxi->xi_win_height,
			NULL);

	XtRealizeWidget(gxi->xi_win);

	cattr.background_pixel = BlackPixel(XtDisplay(gxi->xi_win), 0);
	XChangeWindowAttributes(XtDisplay(gxi->xi_win),
				XtWindow(gxi->xi_win),
				CWBackPixel,
				&cattr);

	if (!gxi->xi_dgainfo->dga_mode)
		xv_init(gvi, gxi);

	XSelectInput(XtDisplay(gxi->xi_win),
			DefaultRootWindow(XtDisplay(gxi->xi_win)),
			SubstructureNotifyMask);

	if (gxi->xi_dgainfo->dga_mode) {
	    for (;;) {
restart_dga_eloop:
		XtAppNextEvent(gxi->xi_app, &e);
		if (e.xany.window == XtWindow(gxi->xi_win)) {
			XtDispatchEvent(&e);
		} else if (e.xany.window != XtWindow(gxi->xi_win) &&
			   e.type == ConfigureNotify && 
			   gxi->xi_wm_parent != 0) {

			/* reconfig onyly if window is in clist & 
			 * pos./dim. change or 
			 * window's rect is within our video win's rect
			 */

			/* Clist, Position & Dimension check */
			for (i = 0; i < gxi->xi_set_clist_count; ++i) {
				if (e.xconfigure.window ==
					gxi->xi_set_clist[i].ci_a.win) {
				    wattr.width = 0;
				    eh = XSetErrorHandler(
						xgetwinattr_errorhandler);
				    XGetWindowAttributes(XtDisplay(gxi->xi_win),
							 e.xconfigure.window,
							 &wattr);
				    XSetErrorHandler(eh);
				    if (wattr.width == 0)
					continue;

				    if (!(wattr.map_state & IsViewable) ||
					wattr.x != gxi->xi_set_clist[i].ci_ox ||
					wattr.y != gxi->xi_set_clist[i].ci_oy ||
					wattr.width != 
						gxi->xi_set_clist[i].ci_ow ||
					wattr.height != 
						gxi->xi_set_clist[i].ci_oh
				    ) {
					    DMSG("Pos -> clip_reconfig\n");
					    clipping_reconfigure(gvi, gxi, 0,0);
					    if (gxi->xi_win_mapped == 1 && 
						gvi->vi_flag_on == 0) {
							start_video(gvi, gxi);
					    }
				    }
				    goto restart_dga_eloop;
				}
			}

			/*RectInRegion check */
			clipping_reconfigure(gvi, gxi, e.xconfigure.window, &i);
			if (i != RectangleOut) {
				DMSG("InRect -> clip_reconfig\n");
				clipping_reconfigure(gvi, gxi, 0, 0);
				if (gxi->xi_win_mapped == 1 && 
					gvi->vi_flag_on == 0) {
						start_video(gvi, gxi);
			        }
			}
		}
	    }
	} else {
	    /* Xv Mode */
	    XtAppMainLoop(gxi->xi_app);
	}
	return 0;
}


static void
alloc_defaults(struct vinfo **vi, struct ainfo **ai, struct xinfo **xi)
{
	if ( (*vi = (struct vinfo*) calloc(1, sizeof(struct vinfo))) == NULL ||
	     (*ai = (struct ainfo*) calloc(1, sizeof(struct ainfo))) == NULL ||
	     (*xi = (struct xinfo*) calloc(1, sizeof(struct xinfo))) == NULL)
		err(ENOMEM, "alloc_defaults");

	if (((*xi)->xi_dgainfo = (struct dgainfo*) 
				calloc(1, sizeof(struct dgainfo))) == NULL ||
	     ((*xi)->xi_xvinfo =  (struct xvinfo*)
				calloc(1, sizeof(struct xvinfo))) == NULL)
		err(ENOMEM, "alloc_defaults");

	/* VI Defaults */
	(*vi)->vi_vfmt = VFMT_YUV422;
	(*vi)->vi_vfmt_swapbytes = SWAPBYTES_2143;
	(*vi)->vi_dname = "/dev/saa0";

	/* AI Defaults */
	(*ai)->ai_ocs = OCS_INLEFT2_INRIGHT2;
	(*ai)->ai_dname = "/dev/sau0";

	/* X Defaults */
	(*xi)->xi_win_width = 720;
	(*xi)->xi_win_height = 576;
	(*xi)->xi_win_resetpending = 1;
	

	/* DGA Defaults */
	(*xi)->xi_dgainfo->dga_mode = 0;

	/* Xv Defaults */
	(*xi)->xi_xvinfo->xv_fourcc = FOURCC_YUY2;
	(*xi)->xi_xvinfo->xv_mode = 'w';
	(*xi)->xi_xvinfo->xv_nodshm = 1;
	(*xi)->xi_xvinfo->xv_multibuf = 0;
}


static void
parse_args(int argc, char **argv, 
		struct vinfo *vi, struct ainfo *ai, struct xinfo *xi)
{
	int ch;

	while ( (ch = getopt(argc, argv, optstr)) != -1) {
		switch(ch) {
		case 'M':
			xi->xi_xvinfo->xv_multibuf = 1;
			break;
		case 'X':
			xi->xi_xvinfo->xv_nodshm = 0;
			break;
		case 'f':
			vi->vi_dname = optarg;
			break;
		case 'D':
			switch (optarg[0]) {
			case 'w':
			case 'b':
			case 'o':
			case 'e':
				xi->xi_dgainfo->dga_mode = optarg[0];
				break;
			default:
				errx(ENOTSUP, 
					"Dga mode %c not supported/implemented"
					"\nUsage:\n%s",
					optarg[0], usage);
			}
			break;
		case 'V':
			switch (optarg[0]) {
			case 'w':
			case 'r':
			case 'b':
			case 'o':
			case 'e':
				xi->xi_dgainfo->dga_mode = 0;
				xi->xi_xvinfo->xv_mode = optarg[0];
				break;
			default:
				errx(ENOTSUP, 
					"Xv mode %c not supported\nUsage:\n%s",
					optarg[0], usage);
			}
			break;
		case 'F':
			switch (optarg[0]) {
			case 'u':
				xv_setmodeinfo(FOURCC_YUY2, vi, xi);
				break;
			case 'y':
				xv_setmodeinfo(FOURCC_YV12, vi, xi);
				break;
			case 'r':
				xv_setmodeinfo(FOURCC_RGB, vi, xi);
				break;
			default:
				errx(ENOTSUP, 
					"Xv format %c not supported/implemented"
					"\nUsage:\n%s",
					optarg[0], usage);
			}
			break;
		case 'a':
			switch(optarg[0]) {
			case 'd':
				ai->ai_ocs = OCS_DACLEFT_DACRIGHT;
				break;
			case '1':
				ai->ai_ocs = OCS_INLEFT1_INRIGHT1;
				break;
			case '2':
				ai->ai_ocs = OCS_INLEFT2_INRIGHT2;
				break;
			case '3':
				ai->ai_ocs = OCS_INLEFT1_INLEFT1;
				break;
			case '4':
				ai->ai_ocs = OCS_INRIGHT1_INRIGHT1;
				break;
			case '5':
				ai->ai_ocs = OCS_INLEFT2_INLEFT2;
				break;
			case '6':
				ai->ai_ocs = OCS_INRIGHT2_INRIGHT2;
				break;
			default:
				errx(ENOTSUP, 
					"Invalid audio Lineout, "
					" mode %c not supported"
					"\nUsage:\n%s",
					optarg[0], usage);
			}
			ai->ai_ocs_initflag = 1;
			break;
		case 'A':
			ai->ai_dname = optarg;
			break;
		default:
			errx(ENOTSUP, "Invalid argument %c\nUsage:\n%s",
				ch, usage);
			break;
		}
	}
}


static __inline void
xv_setmodeinfo(uint32_t fourcc, struct vinfo *vi, struct xinfo *xi)
{
	switch (fourcc) {
		case FOURCC_YUY2:
			xi->xi_xvinfo->xv_fourcc = FOURCC_YUY2;
			vi->vi_vfmt = VFMT_YUV422;
			vi->vi_vfmt_swapbytes = SWAPBYTES_2143;
			break;
		case FOURCC_YV12:
			xi->xi_xvinfo->xv_fourcc = FOURCC_YV12;
			vi->vi_vfmt = VFMT_PLANAR_YUV422;
			vi->vi_vfmt_swapbytes = SWAPBYTES_NONE;
			break;
		case FOURCC_RGB:
			xi->xi_xvinfo->xv_fourcc = FOURCC_RGB;
			vi->vi_vfmt = VFMT_RGBa888;
			vi->vi_vfmt_swapbytes = SWAPBYTES_NONE;
			break;
		default:
			errx(ENOTSUP, 
				"Xv fourcc %d not supported/implemented",
				fourcc);
		}
}


static void
audio_init(struct ainfo *ai)
{
	/* Open audio device only if ocs is specified */
	if (ai->ai_ocs_initflag != 0) {
		if ( (ai->ai_dfd = open(ai->ai_dname, O_RDWR)) == -1)
			warn("error opening %s", ai->ai_dname);
		else 
			if (ioctl(ai->ai_dfd, SETAUDIOMIXER, &ai->ai_ocs) == -1)
                		err(errno, "Audio set lineout, Ioctl error");
	}
}


static void
dga_init(struct vinfo *vi, struct xinfo *xi)
{
	Display *d;
	int depth, s, dummy1, dummy2;

	d = XtDisplay(xi->xi_win);
	s = DefaultScreen(XtDisplay(xi->xi_win));
	if (!XDGAQueryVersion(d, &dummy1, &dummy2) || dummy1 < 2)
                errx(ENOTSUP, 
			"Program require XF86DGA, version 2.0 or greater");
        XF86DGAGetVideoLL(d, s, 
				(int*)&xi->xi_dgainfo->dga_base, 
				&xi->xi_dgainfo->dga_ppl, 
				&dummy1, &dummy2);
	depth = DefaultDepth(d, s);
	/* stopgap till we do fbpp detection */
	if (depth == 24)
		depth = 32;
	xi->xi_dgainfo->dga_bypp = (depth + 1) / 8;
        xi->xi_dgainfo->dga_flags = 0;
	switch (depth) {
                case 32:
                case 24:
                        vi->vi_vfmt = VFMT_RGBa888;
                        break;
                case 16:
                        vi->vi_vfmt = VFMT_RGB565;
                        xi->xi_dgainfo->dga_flags |= VSPEC_FLAG_DITHER;
                        break;
                case 15:
                        vi->vi_vfmt = VFMT_RGBa555;
                        xi->xi_dgainfo->dga_flags |= VSPEC_FLAG_DITHER;
                        break;
		case 8:
                        vi->vi_vfmt = VFMT_Y8;
                        xi->xi_dgainfo->dga_flags |= VSPEC_FLAG_DITHER;
                        break;
                default:
                        errx(ENOTSUP, 
				"error programe require desktop depth "
				"to be greater than or equal to 15,\n"
				"current depth: %d", depth);
                        break;
        }
}


static void
xv_init(struct vinfo *vi, struct xinfo *xi)
{
	struct sigaction sa;
	sigset_t ss;
	Display *d;
	XvAdaptorInfo *adp;
	uint32_t i, j;
	
	sa.sa_flags = 0;
	sigemptyset(&ss);
	sa.sa_mask = ss;
	sa.sa_handler = terminate_notify;
	if (sigaction(SIGINT, &sa, NULL) != 0)
		err(errno, "error setting up status signal handler (INT)");
	if (sigaction(SIGHUP, &sa, NULL) != 0)
		err(errno, "error setting up status signal handler (HUP)");

	if (xi->xi_xvinfo->xv_multibuf) {
		if (xi->xi_xvinfo->xv_nodshm == 0)
			err(EINVAL, 
				"Command line error, MULTIBUF (-M) and "
				"DSHM (-X) are mutually exclusive options");
		sa.sa_handler = xv_multibuf_notify;
	} else {
		sa.sa_handler = xv_frame_notify;
	}
	if (sigaction(SIGUSR1, &sa, NULL) != 0)
		err(errno, "error setting up frame signal handler (USR1)");

	sa.sa_handler = xv_status_notify;
	if (sigaction(SIGUSR2, &sa, NULL) != 0)
		err(errno, "error setting up status signal handler (USR2)");
	if ( (vi->vi_dsts = 
		(struct saa_status*) mmap(0, SAA_MMAP_STS_SZ, PROT_READ, 0,
				vi->vi_dfd, SAA_MMAP_STS_OFFSET)) == MAP_FAILED)
		err(errno, "error mapping status buffer");
	d = XtDisplay(xi->xi_win);
	if (XvQueryAdaptors(d,
				XtWindow(xi->xi_win),
				&i, &adp) != Success)
		errx(EIO, "error querying Xv adaptors");
	if (i == 0)
		errx(EIO, "error no Xv adaptors present");
	j = adp[0].base_id + adp[0].num_ports;
	for (i = adp[0].base_id; i < j; ++i)
		if (XvGrabPort(d, i, CurrentTime) == Success)
			goto l1;
	errx(EIO, "error no Xv ports free");
l1:
	xi->xi_xvinfo->xv_port = i;
	XvFreeAdaptorInfo(adp);
	xi->xi_xvinfo->xv_gc = XCreateGC(d, XtWindow(xi->xi_win), 0, NULL);
}


static void
xv_status_notify(int sig)
{
	DMSG("In status notify AVSTS: 0x%x IREP: 0x%x\n", 
		gvi->vi_dsts->ssts_video_sts,
		gvi->vi_dsts->ssts_irep);
}


static void
xv_multibuf_notify(int sig)
{
	struct saa_status *ssts;
	uint32_t offset;
	
	ssts = gvi->vi_dsts;
	if ((ssts->ssts_video_framectr - gxi->xi_xvinfo->xv_mbuf_ctr) >=
					(ssts->ssts_video_mb_count - 1)) 
	{
		DMSG("Video drop: realigning\n");
		gxi->xi_xvinfo->xv_mbuf_ctr = ssts->ssts_video_framectr - 2;
	}
	offset = gxi->xi_xvinfo->xv_mbuf_ctr % ssts->ssts_video_mb_count;
	offset *= ssts->ssts_video_mbufsz;
	++gxi->xi_xvinfo->xv_mbuf_ctr;

	if (gxi->xi_xvinfo->xv_nodshm) {
#ifdef TVV_USE_MMXEXT
	    special_memcpy_mmxext(gxi->xi_xvinfo->xv_shmaddr, 
					gxi->xi_xvinfo->xv_mmapaddr + offset, 
					gxi->xi_xvinfo->xv_framesz);
#else
	    special_memcpy(gxi->xi_xvinfo->xv_shmaddr, 
					gxi->xi_xvinfo->xv_mmapaddr + offset, 
					gxi->xi_xvinfo->xv_framesz);
#endif
	}

	/* INVARIANT multibuf not allowed in DSHM mode */

	if (XvShmPutImage(XtDisplay(gxi->xi_win), 
			gxi->xi_xvinfo->xv_port, 
			XtWindow(gxi->xi_win), 
			gxi->xi_xvinfo->xv_gc,
			gxi->xi_xvinfo->xv_img,
			0, 0,
			gvi->vi_svs.svs_buf_width,gxi->xi_xvinfo->xv_img_height,
			0, 0,
			gxi->xi_win_width, gxi->xi_win_height,
			False) != Success)
		warn("Xv put image failed");
	XSync(XtDisplay(gxi->xi_win), False);
	if (gvi->vi_dsts->ssts_sig_sts)
		xv_status_notify(sig);
}


static void
xv_frame_notify(int sig)
{
	if (gxi->xi_xvinfo->xv_nodshm) {
#ifdef TVV_USE_MMXEXT
	    special_memcpy_mmxext(gxi->xi_xvinfo->xv_shmaddr, 
					gxi->xi_xvinfo->xv_mmapaddr, 
					gxi->xi_xvinfo->xv_framesz);
#else
	    special_memcpy(gxi->xi_xvinfo->xv_shmaddr, 
					gxi->xi_xvinfo->xv_mmapaddr, 
					gxi->xi_xvinfo->xv_framesz);
#endif
	}

	if (XvShmPutImage(XtDisplay(gxi->xi_win), 
			gxi->xi_xvinfo->xv_port, 
			XtWindow(gxi->xi_win), 
			gxi->xi_xvinfo->xv_gc,
			gxi->xi_xvinfo->xv_img,
			0, 0,
			gvi->vi_svs.svs_buf_width,gxi->xi_xvinfo->xv_img_height,
			0, 0,
			gxi->xi_win_width, gxi->xi_win_height,
			False) != Success)
		warn("Xv put image failed");
	XSync(XtDisplay(gxi->xi_win), False);
	if (gvi->vi_dsts->ssts_sig_sts)
		xv_status_notify(sig);
}


static void
terminate_notify(int sig)
{
	stop_video(gvi, gxi);
	if (gxi->xi_xvinfo->xv_shmaddr) {
		if (gxi->xi_xvinfo->xv_img) {  
			XFree(gxi->xi_xvinfo->xv_img);
			gxi->xi_xvinfo->xv_img = 0;
		}
		XShmDetach(XtDisplay(gxi->xi_win),&gxi->xi_xvinfo->xv_xshminfo);
		if (shmdt(gxi->xi_xvinfo->xv_shmaddr) == -1)
			warn("Error detaching shared memory");
		gxi->xi_xvinfo->xv_shmaddr = 0;
		shmctl(gxi->xi_xvinfo->xv_shmid, IPC_RMID, 0);
	}
	exit(0);
}


static void
start_video(struct vinfo *vi, struct xinfo *xi)
{
	DMSG("Starting Video\n");
	if (xi->xi_dgainfo->dga_mode) {
		generate_saa_cliplist(xi, &vi->vi_svs.svs_cliplist);
		if (ioctl(vi->vi_dfd, OVERLAYON, &vi->vi_svs) != 0)
			warn("DGA start video failed");
		else
			vi->vi_flag_on = 1;
	} else {
		if (ioctl(vi->vi_dfd, CAPTUREON, &vi->vi_svs) != 0)
			warn("Device start video failed");
		else {
			vi->vi_flag_on = 1;
			if (xi->xi_xvinfo->xv_nodshm) {
				if (xi->xi_xvinfo->xv_mmapaddr == 0)
					xi->xi_xvinfo->xv_mmapaddr = (uint8_t*)
						mmap(0, 
							SAA_VCAP_FRAME_MAX_SZ,
							PROT_READ, 0,
							vi->vi_dfd, 
							SAA_MMAP_T0_OFFSET); 
				if (xi->xi_xvinfo->xv_mmapaddr == MAP_FAILED)
				    err(errno, 
					"Error mapping device frame buffer");
			}
		}
	}
}


static void
stop_video(struct vinfo *vi, struct xinfo *ai)
{
	int dummy;
	if (gxi->xi_dgainfo->dga_mode) {
		if (ioctl(gvi->vi_dfd, OVERLAYOFF, &dummy) != 0)
			warn("DGA stop_video failed");
	} else {

/*
 * XXX Not required due to saa device buffer allocation strategy??
 *     Add to atexit dnit code
 */
#if XXX_NOTREQUIRED
		if (xi->xi_xvinfo->xv_nodshm) {
			if (xi->xi_xvinfo->xv_mmaplen) {
			    	if (munmap(xi->xi_xvinfo->xv_mmapaddr, 
					SAA_VCAP_FRAME_MAX_SZ)!= 0)
					warn("munmap failure, mem leak!\n");
			}
		}
#endif
		if (ioctl(vi->vi_dfd, CAPTUREOFF, &dummy) != 0)
			warn("Xv stop_video failed");
	}
	DMSG("Stopping Video\n");
	vi->vi_flag_on = 0;
}


static void
reset_viewport(struct vinfo *vi, struct xinfo *xi)
{
	int i, bpp;
	short x, y, width, height;

	stop_video(vi, xi);
	if (xi->xi_dgainfo->dga_mode) {
		x = xi->xi_win_xpos;
		y = xi->xi_win_ypos;
		width = xi->xi_win_width;
		height = xi->xi_win_height;
		/* XXX use clipping instead, when done */
#if XXX_HANDLED_BY_CLIPPING
		if (x < 0) {
			width += x;
			x = 0;
			vi->vi_svs.svs_flags |= VSPEC_FLAG_NO_HSCALE;
		}
		if (y < 0) {
			height += y;
			y = 0;
			vi->vi_svs.svs_flags |= VSPEC_FLAG_NO_VSCALE;
		}
#endif
		vi->vi_svs.svs_buf_base = (char*)xi->xi_dgainfo->dga_base;
		vi->vi_svs.svs_buf_base += 
		    (y * (xi->xi_dgainfo->dga_ppl * xi->xi_dgainfo->dga_bypp)) +
		    (x * xi->xi_dgainfo->dga_bypp);

		if (width + x >= xi->xi_scr_width) {
			vi->vi_svs.svs_buf_width = xi->xi_scr_width - x;
			vi->vi_svs.svs_flags |= VSPEC_FLAG_NO_HSCALE;
		} else
			vi->vi_svs.svs_buf_width = width;

		if (height + y >= xi->xi_scr_height) {
			vi->vi_svs.svs_buf_height = xi->xi_scr_height - y;
			vi->vi_svs.svs_flags |= VSPEC_FLAG_NO_VSCALE;
		} else
			vi->vi_svs.svs_buf_height = height;

		vi->vi_svs.svs_vfmt = (enum VIDEOFMT)vi->vi_vfmt;
		vi->vi_svs.svs_flags |= 
			xi->xi_dgainfo->dga_flags | VSPEC_FLAG_FRAMELOCK;
		switch (xi->xi_dgainfo->dga_mode) {
		case 'w':
			/* Weave-RT mode 50 ups */
			vi->vi_svs.svs_buf_offset[0][0] = 0;
			vi->vi_svs.svs_buf_offset[1][0] = 
			    xi->xi_dgainfo->dga_ppl * xi->xi_dgainfo->dga_bypp;
			vi->vi_svs.svs_buf_pitch[0] = xi->xi_dgainfo->dga_ppl *
						xi->xi_dgainfo->dga_bypp * 2;
			break;
		case 'b':
			/* Bob mode 50 ups */
			vi->vi_svs.svs_buf_offset[0][0] = 0;
			vi->vi_svs.svs_buf_offset[1][0] = 0;
			vi->vi_svs.svs_buf_pitch[0] = xi->xi_dgainfo->dga_ppl *
						xi->xi_dgainfo->dga_bypp;
			vi->vi_svs.svs_buf_height *= 2;
			break;
		case 'o':
			/* Drop odd field mode 25 ups */
			vi->vi_svs.svs_buf_offset[0][0] = 0;
			vi->vi_svs.svs_buf_offset[1][0] = 0;
			vi->vi_svs.svs_buf_pitch[0] =
			    xi->xi_dgainfo->dga_ppl * xi->xi_dgainfo->dga_bypp;
			vi->vi_svs.svs_buf_height *= 2;
			vi->vi_svs.svs_flags |= VSPEC_FLAG_OVERLAY_DROPODD;
			break;
		case 'e':
			/* Drop even field mode 25 ups */
			vi->vi_svs.svs_buf_offset[0][0] = 0;
			vi->vi_svs.svs_buf_offset[1][0] = 0;
			vi->vi_svs.svs_buf_pitch[0] =
			    xi->xi_dgainfo->dga_ppl * xi->xi_dgainfo->dga_bypp;
			vi->vi_svs.svs_buf_height *= 2;
			vi->vi_svs.svs_flags |= VSPEC_FLAG_OVERLAY_DROPEVEN;
			break;
		}
	} else {
		/* Xv Mode */
		vi->vi_svs.svs_buf_width = xi->xi_win_width < vi->vi_src_width ?
					xi->xi_win_width : vi->vi_src_width;
		vi->vi_svs.svs_buf_height = xi->xi_win_height < vi->vi_src_height ?
					xi->xi_win_height : 
					vi->vi_src_height;
		vi->vi_svs.svs_buf_width = 
			((vi->vi_svs.svs_buf_width + 15) / 16) * 16;
		vi->vi_svs.svs_buf_height = 
			((vi->vi_svs.svs_buf_height + 15) / 16) * 16;
		if (xi->xi_xvinfo->xv_shmaddr) {
			/* XXX Memory Leak?, figure out how to delete XvImage */
			if (xi->xi_xvinfo->xv_img) {  
				XFree(xi->xi_xvinfo->xv_img);
				xi->xi_xvinfo->xv_img = 0;
			}
			XShmDetach(XtDisplay(xi->xi_win), 
					&xi->xi_xvinfo->xv_xshminfo);
			if (shmdt(xi->xi_xvinfo->xv_shmaddr) == -1)
				warn("error detaching shared memory");
			xi->xi_xvinfo->xv_shmaddr = 0;
			shmctl(xi->xi_xvinfo->xv_shmid, IPC_RMID, 0);
		}
		i = SHM_R | (SHM_R >> 3) | (SHM_R >> 6);
		i |= SHM_W | (SHM_W >> 3) | (SHM_W >> 6);
		xi->xi_xvinfo->xv_shmid = shmget(TVV_SHMKEY, 
							SAA_VCAP_FRAME_MAX_SZ, 
							i);
		if (xi->xi_xvinfo->xv_shmid != -1) {
			if (shmctl(xi->xi_xvinfo->xv_shmid, IPC_RMID, 0) == -1)
				warnx("error removing stale shared memory "
					"segment, id: 0x%x", 
					xi->xi_xvinfo->xv_shmid);
		}
		i |= IPC_CREAT;
		xi->xi_xvinfo->xv_shmid = shmget(TVV_SHMKEY, 
							SAA_VCAP_FRAME_MAX_SZ,
							i);
		if (xi->xi_xvinfo->xv_shmid == -1)
			err(errno, "error getting shared memory");
		xi->xi_xvinfo->xv_shmaddr = 
			(uint8_t*) shmat(xi->xi_xvinfo->xv_shmid, 0, 0);
		if (xi->xi_xvinfo->xv_shmaddr == (void*)-1)
			err(errno, "error attaching to shared memory");
#if XXX_DISABLED_FOR_HWPIP
		bzero(xi->xi_xvinfo->xv_shmaddr, SAA_VCAP_FRAME_MAX_SZ);
#endif
		xi->xi_xvinfo->xv_xshminfo.shmid = xi->xi_xvinfo->xv_shmid;
		xi->xi_xvinfo->xv_xshminfo.shmaddr = 
				(char*) xi->xi_xvinfo->xv_shmaddr;
		xi->xi_xvinfo->xv_xshminfo.readOnly = False;
		XShmAttach(XtDisplay(xi->xi_win), &xi->xi_xvinfo->xv_xshminfo);
		xi->xi_xvinfo->xv_img_height = vi->vi_svs.svs_buf_height;
		/* XXX Weaving not possible with VFMT_PLANAR_YUV420 */
		if (xi->xi_xvinfo->xv_mode == 'b' || 
			xi->xi_xvinfo->xv_mode == 'o' || 
			xi->xi_xvinfo->xv_mode == 'e') {
			xi->xi_xvinfo->xv_img_height /= 2;
			if (xi->xi_xvinfo->xv_fourcc == FOURCC_YV12)
				vi->vi_vfmt = VFMT_PLANAR_YUV420;
		} else {
			if (xi->xi_xvinfo->xv_fourcc == FOURCC_YV12)
				vi->vi_vfmt = VFMT_PLANAR_YUV422;
		}
		xi->xi_xvinfo->xv_img = XvShmCreateImage(XtDisplay(xi->xi_win), 
					xi->xi_xvinfo->xv_port,
					xi->xi_xvinfo->xv_fourcc,
					0,
					vi->vi_svs.svs_buf_width, 
					xi->xi_xvinfo->xv_img_height,
					&xi->xi_xvinfo->xv_xshminfo);
		if (!xi->xi_xvinfo->xv_img)
			errx(EIO, "error creating Xv image");
		xi->xi_xvinfo->xv_img->data = (char*)xi->xi_xvinfo->xv_shmaddr;

		vi->vi_svs.svs_cap_nrtasks = 1;
		vi->vi_svs.svs_cap_shmid[0] = xi->xi_xvinfo->xv_shmid;
		vi->vi_svs.svs_vfmt = (enum VIDEOFMT)vi->vi_vfmt;
		vi->vi_svs.svs_cap_swapbytes = vi->vi_vfmt_swapbytes;
		bpp = (vi->vi_vfmt & VFMT_BPP_MASK) >> VFMT_BPP_SHIFT;
		xi->xi_xvinfo->xv_framesz = 
			(xi->xi_xvinfo->xv_img_height * 
					vi->vi_svs.svs_buf_width * bpp) / 8;
		xi->xi_xvinfo->xv_framesz = 
				((xi->xi_xvinfo->xv_framesz + 63) / 64) * 64;
		/*vi->vi_svs.svs_flags == 0, due to decl. as global var. */

		switch (xi->xi_xvinfo->xv_mode) {
		case 'w':	/* Weave */
		case 'r':
			if (xi->xi_xvinfo->xv_fourcc == FOURCC_YUY2 || 
			    xi->xi_xvinfo->xv_fourcc == FOURCC_RGB) {
				vi->vi_svs.svs_buf_offset[0][0] = 0;
				vi->vi_svs.svs_buf_offset[1][0] = 
					(vi->vi_svs.svs_buf_width * bpp) / 8;
				vi->vi_svs.svs_buf_pitch[0] = 
				    (vi->vi_svs.svs_buf_width * bpp * 2) / 8;
			} else {
			    vi->vi_svs.svs_buf_offset[0][0] = 0;
			    vi->vi_svs.svs_buf_offset[1][0] = 
					vi->vi_svs.svs_buf_width;
			    vi->vi_svs.svs_buf_pitch[0] = 
					vi->vi_svs.svs_buf_width * 2;

			    vi->vi_svs.svs_buf_offset[0][2] = 
					vi->vi_svs.svs_buf_width * 
						vi->vi_svs.svs_buf_height;
			    vi->vi_svs.svs_buf_offset[1][2] = 
				vi->vi_svs.svs_buf_offset[0][2] + 
				(vi->vi_svs.svs_buf_width * 
					vi->vi_svs.svs_buf_height / 2);
			    vi->vi_svs.svs_buf_pitch[2] = 
					vi->vi_svs.svs_buf_width / 2;

			    vi->vi_svs.svs_buf_offset[0][1] = 
				vi->vi_svs.svs_buf_width * 
				vi->vi_svs.svs_buf_height +
			 	(vi->vi_svs.svs_buf_width * 
					vi->vi_svs.svs_buf_height / 4);
			    vi->vi_svs.svs_buf_offset[1][1] = 
				vi->vi_svs.svs_buf_offset[0][1] + 
				(vi->vi_svs.svs_buf_width * 
					vi->vi_svs.svs_buf_height / 2);
			    vi->vi_svs.svs_buf_pitch[1] = 
					vi->vi_svs.svs_buf_width / 2;
			}
			if (xi->xi_xvinfo->xv_mode == 'r')	/* Weave-RT */
				vi->vi_svs.svs_flags |= 
					VSPEC_FLAG_CAPTURE_SIGFLD;
			break;
		case 'b':
			if (xi->xi_xvinfo->xv_fourcc == FOURCC_YUY2 || 
			    xi->xi_xvinfo->xv_fourcc == FOURCC_RGB) {
				vi->vi_svs.svs_buf_offset[0][0] = 0;
				vi->vi_svs.svs_buf_offset[1][0] = 0;
				vi->vi_svs.svs_buf_pitch[0] = 
					(vi->vi_svs.svs_buf_width * bpp) / 8;
			} else {
			    vi->vi_svs.svs_buf_offset[0][0] = 0;
			    vi->vi_svs.svs_buf_offset[1][0] = 0;
			    vi->vi_svs.svs_buf_pitch[0] = 
					vi->vi_svs.svs_buf_width;

			    vi->vi_svs.svs_buf_offset[0][2] = 
				(vi->vi_svs.svs_buf_width * 
					xi->xi_xvinfo->xv_img_height);
			    vi->vi_svs.svs_buf_offset[1][2] = 
				vi->vi_svs.svs_buf_offset[0][2];
			    vi->vi_svs.svs_buf_pitch[2] = 
				vi->vi_svs.svs_buf_width / 2;

			    vi->vi_svs.svs_buf_offset[0][1] = 
				(vi->vi_svs.svs_buf_width * 
					xi->xi_xvinfo->xv_img_height) +
			 	((vi->vi_svs.svs_buf_width * 
					xi->xi_xvinfo->xv_img_height) / 4);
			    vi->vi_svs.svs_buf_offset[1][1] = 
					vi->vi_svs.svs_buf_offset[0][1];
			    vi->vi_svs.svs_buf_pitch[1] = 
					vi->vi_svs.svs_buf_width / 2;
			}
			vi->vi_svs.svs_flags |= VSPEC_FLAG_CAPTURE_SIGFLD;
			break;
		case 'o':
		case 'e':
			if (xi->xi_xvinfo->xv_fourcc == FOURCC_YUY2 || 
			    xi->xi_xvinfo->xv_fourcc == FOURCC_RGB) {
				vi->vi_svs.svs_buf_offset[0][0] = 0;
				vi->vi_svs.svs_buf_pitch[0] = 
					(vi->vi_svs.svs_buf_width * bpp) / 8;
				vi->vi_svs.svs_buf_offset[1][0] = 
					(vi->vi_svs.svs_buf_pitch[0] * 
						vi->vi_svs.svs_buf_height) / 2;
				if (xi->xi_xvinfo->xv_mode == 'e') {
					vi->vi_svs.svs_buf_offset[0][0] = 
						vi->vi_svs.svs_buf_offset[0][1];
					vi->vi_svs.svs_buf_offset[0][1] = 0;
				}
			} else {
			    i = (vi->vi_svs.svs_buf_width * 
				xi->xi_xvinfo->xv_img_height * 3) / 2;
			    vi->vi_svs.svs_buf_offset[0][0] = 0;
			    vi->vi_svs.svs_buf_offset[1][0] = i;
			    vi->vi_svs.svs_buf_pitch[0] = 
				vi->vi_svs.svs_buf_width;

			    vi->vi_svs.svs_buf_offset[0][2] = 
				(vi->vi_svs.svs_buf_width * 
					xi->xi_xvinfo->xv_img_height);
			    vi->vi_svs.svs_buf_offset[1][2] = 
				vi->vi_svs.svs_buf_offset[0][2] + i;
			    vi->vi_svs.svs_buf_pitch[2] = 
				vi->vi_svs.svs_buf_width / 2;

			    vi->vi_svs.svs_buf_offset[0][1] = 
				(vi->vi_svs.svs_buf_width * 
					xi->xi_xvinfo->xv_img_height) +
			 	((vi->vi_svs.svs_buf_width * 
					xi->xi_xvinfo->xv_img_height) / 4);
			    vi->vi_svs.svs_buf_offset[1][1] = 
				vi->vi_svs.svs_buf_offset[0][1] + i;
			    vi->vi_svs.svs_buf_pitch[1] = 
					vi->vi_svs.svs_buf_width / 2;
			    if (xi->xi_xvinfo->xv_mode == 'e') {
				i = vi->vi_svs.svs_buf_offset[0][0];
				vi->vi_svs.svs_buf_offset[0][0] = 
					vi->vi_svs.svs_buf_offset[1][0];
				vi->vi_svs.svs_buf_offset[1][0] = i;

				i = vi->vi_svs.svs_buf_offset[0][1];
				vi->vi_svs.svs_buf_offset[0][1] = 
					vi->vi_svs.svs_buf_offset[1][1];
				vi->vi_svs.svs_buf_offset[1][1] = i;

				i = vi->vi_svs.svs_buf_offset[0][2];
				vi->vi_svs.svs_buf_offset[0][2] = 
					vi->vi_svs.svs_buf_offset[1][2];
				vi->vi_svs.svs_buf_offset[1][2] = i;
			    }
			}
			break;
		}
		vi->vi_svs.svs_flags |= VSPEC_FLAG_CAPTURE_SIGNAL |
					VSPEC_FLAG_STATUS_SIGNAL |
					VSPEC_FLAG_FRAMELOCK;
		vi->vi_svs.svs_cap_sigmask = STS_ALL;
		if (xi->xi_xvinfo->xv_nodshm) {
			  vi->vi_svs.svs_cap_bufsz = SAA_VCAP_FRAME_MAX_SZ;
		} else
			vi->vi_svs.svs_flags |= VSPEC_FLAG_CAPTURE_SHM;
		if (xi->xi_xvinfo->xv_multibuf) {
			vi->vi_svs.svs_flags |= VSPEC_FLAG_CAPTURE_MULTIBUF;
		}
	}
}


static void
event_handler(Widget w, XtPointer priv, XEvent *e, Boolean *cont)
{
	struct privdata *gptr;
	struct vinfo *vi;
	struct ainfo *ai;
	struct xinfo *xi;
	unsigned int i;
	short x, y;

	gptr = (struct privdata*) priv;
	vi = gptr->vi; ai = gptr->ai; xi = gptr->xi;

	if (e->xany.window != XtWindow(xi->xi_win)) {
		*cont = TRUE;
		return;
	}

	switch (e->type) {
	case Expose:
		if (!xi->xi_dgainfo->dga_mode)
			break;
		clipping_reconfigure(vi, xi, 0, 0);
		if (xi->xi_win_mapped == 1 && vi->vi_flag_on == 0) 
			start_video(vi, xi);
		break;
	case VisibilityNotify:
		if (!xi->xi_dgainfo->dga_mode)
			break;
		switch (e->xvisibility.state) {
		case VisibilityUnobscured:
			clipping_reset(vi, xi);
			if (xi->xi_win_mapped == 1 && vi->vi_flag_on == 0) 
				start_video(vi, xi);
			break;
		case VisibilityFullyObscured:
			if (xi->xi_dgainfo->dga_mode)
				stop_video(vi, xi);
			break;
		case VisibilityPartiallyObscured:
			clipping_reconfigure(vi, xi, 0, 0);
			if (xi->xi_win_mapped == 1 && vi->vi_flag_on == 0)
				start_video(vi, xi);
			break;
		}
		break;
	case UnmapNotify:
		xi->xi_win_mapped = 0;
		stop_video(vi, xi);
		if (!xi->xi_dgainfo->dga_mode)
			xi->xi_win_resetpending = 1;
		break;
	case MapNotify:
		xi->xi_win_mapped = 1;
		if (xi->xi_win_resetpending == 1) {
			xi->xi_win_resetpending = 0;
			reset_viewport(vi, xi);
		}
		start_video(vi, xi);
		break;
	case ConfigureNotify:
		i = 0;
		/* XXX cannot rely on xconfigure.[xy] as
		 * they become 0, 0 when win is resized
		 */
		XtVaGetValues(xi->xi_win,
				XtNx, &x,	
				XtNy, &y,	
				NULL);
		if (xi->xi_win_width != e->xconfigure.width) {
			if (!xi->xi_dgainfo->dga_mode && 
				(xi->xi_win_width >= vi->vi_src_width &&
				 e->xconfigure.width >= vi->vi_src_width))
				xi->xi_win_width = e->xconfigure.width;
			else {
				xi->xi_win_width = e->xconfigure.width;
				++i;
			}
		}
		if (xi->xi_win_height != e->xconfigure.height) {
			if (!xi->xi_dgainfo->dga_mode && 
				(xi->xi_win_height >= vi->vi_src_height &&
				 e->xconfigure.height >= vi->vi_src_height))
				xi->xi_win_height = e->xconfigure.height;
			else {
				xi->xi_win_height = e->xconfigure.height;
				++i;
			}
		}
		if (xi->xi_win_xpos != x) {
			xi->xi_win_xpos = x;
			if (xi->xi_dgainfo->dga_mode)
				++i;
		}
		if (xi->xi_win_ypos != y) {
			xi->xi_win_ypos = y;
			if (xi->xi_dgainfo->dga_mode)
				++i;
		}
		if (i) {
			xi->xi_win_resetpending = 1;
			if (xi->xi_win_mapped) {
				xi->xi_win_resetpending = 0;
				clipping_reconfigure(vi, xi, 0, 0);
				reset_viewport(vi, xi);
				start_video(vi, xi);
			}
			if (xi->xi_dgainfo->dga_mode &&
			    xi->xi_dgainfo->dga_wproc_id == 0)
			{
			    xi->xi_dgainfo->dga_wproc_id =
				XtAppAddWorkProc(xi->xi_app, refresh_desktop, xi);
			}
		}
		break;
	}
	*cont = False;
}


int
xgetwinattr_errorhandler(Display *d, XErrorEvent *e) 
{
	return TRUE;
}


static void
get_source_dimensions(struct vinfo *vi, int *width, int *height)
{
	struct saa_video_std vstd;

	if (ioctl(vi->vi_dfd, GETVIDEOSTD, &vstd) != 0)
		errx(errno, "error getting video standard");

	*width = vstd.std_video_xinstop -
			vstd.std_video_xinstart + 1;
	*height = (vstd.std_video_yinstop -   
			vstd.std_video_yinstart + 1) * 2;

}
