#include <sys/param.h>
#include <errno.h>
#include <err.h>
#include <stdint.h>
#include <stdlib.h>

#include "tvv.h"
#include "tvv_clip.h"

static __inline void cliplist_clear(struct xinfo *xi);
static __inline void cliplist_add(struct xinfo *xi, struct clipinfo *ci);

static int	clipinfo_compare(const void *a, const void *b);
static void	cliplist_sort(struct xinfo *xi);
static void	cliplist_merge(struct xinfo *xi);
static void	clipping_set_cliplist(struct vinfo *vi, struct xinfo *xi);


static __inline void
cliplist_clear(struct xinfo *xi)
{
	xi->xi_clist_count = 0;	
}


static __inline void
cliplist_add(struct xinfo *xi, struct clipinfo *ci)
{
	struct clipinfo *nci;
	int i;

	if (xi->xi_clist_count >= xi->xi_clist_sz) {
		nci = (struct clipinfo*) 
			malloc(sizeof(struct clipinfo) * 
					(xi->xi_clist_sz + CLIPLIST_GROWSZ));
		if (!nci)
			errx(ENOMEM, "error allocating cliplist memory");
		memcpy(nci, xi->xi_clist, 
			sizeof(struct clipinfo) * xi->xi_clist_count);
		free(xi->xi_clist);
		xi->xi_clist = nci;
		xi->xi_clist_sz += CLIPLIST_GROWSZ;
	}
	
	xi->xi_clist[xi->xi_clist_count] = *ci;
	++xi->xi_clist_count;

	DMSG("InClipAdd (0x%x): x: %d y: %d w: %d h: %d\n", 
		 ci->ci_a.win, ci->ci_a.x, ci->ci_a.y, ci->ci_a.w, ci->ci_a.h);
}


static int
clipinfo_compare(const void *a, const void *b)
{
	struct clipinfo *cia, *cib;
	uint32_t i, j;
	
	cia = (struct clipinfo*) a;
	cib = (struct clipinfo*) b;

	i = cia->ci_skip == -1 ? 0 : 1;
	j = cib->ci_skip == -1 ? 0 : 1;

	if (i != j) {
		if (i)
			return 1;
		else
			return -1;
	}

	/* Invariant, i == j */
	if (i == 1)
		return 0;

	i = cia->ci_a.x;
	j = cib->ci_a.x;
	if (i != j) {
		if (i > j)
			return 1;
		else
			return -1;
	}

	i = cia->ci_a.y;
	j = cib->ci_a.y;
	if (i != j) {
		if (i > j)
			return 1;
		else
			return -1;
	}

	i = cia->ci_a.w;
	j = cib->ci_a.w;
	if (i != j) {
		if (i > j)
			return 1;
		else
			return -1;
	}

	i = cia->ci_a.h;
	j = cib->ci_a.h;
	if (i != j) {
		if (i > j)
			return 1;
		else
			return -1;
	}

	/* Regions are identical, should not occur as identical regions
	 *  are merged by clipist_merge
	 */
	return 0;
}


static void
cliplist_sort(struct xinfo *xi)
{
	uint32_t i;

	if (xi->xi_clist_count <= 1)
		return;

	qsort(xi->xi_clist, xi->xi_clist_count, sizeof(struct clipinfo),
		clipinfo_compare);

	for (i = 0; i < xi->xi_clist_count; ++i)
		if (xi->xi_clist[i].ci_skip != -1)
			break;

	xi->xi_clist_count = i;
}


static void
cliplist_merge(struct xinfo *xi)
{
	XPoint points[4];
	Region region;
	uint32_t i, j;

	for (i = 0; i < xi->xi_clist_count; ++i) {

		if (xi->xi_clist[i].ci_skip != -1)
			continue;

		points[0].x = xi->xi_clist[i].ci_a.x;
		points[0].y = xi->xi_clist[i].ci_a.y;
		points[1].x = xi->xi_clist[i].ci_a.x;
		points[1].y = xi->xi_clist[i].ci_a.y + xi->xi_clist[i].ci_a.h;
		points[2].x = xi->xi_clist[i].ci_a.x + xi->xi_clist[i].ci_a.w;
		points[2].y = xi->xi_clist[i].ci_a.y + xi->xi_clist[i].ci_a.h;
		points[3].x = xi->xi_clist[i].ci_a.x + xi->xi_clist[i].ci_a.w;
		points[3].y = xi->xi_clist[i].ci_a.y;
		region = XPolygonRegion(points, 4, WindingRule);

		for (j = 0; j < xi->xi_clist_count; ++j) {
			if (xi->xi_clist[j].ci_skip == -1 && j != i &&
			    (XRectInRegion(region, 
				xi->xi_clist[j].ci_a.x,
				xi->xi_clist[j].ci_a.y,
				xi->xi_clist[j].ci_a.w,
				xi->xi_clist[j].ci_a.h) == RectangleIn)) 
			{
				xi->xi_clist[j].ci_skip = i;
				DMSG("Merging Region %d into %d\n", j, i);
			}
		}
		XDestroyRegion(region);
	}
}


static void
clipping_set_cliplist(struct vinfo *vi, struct xinfo *xi)
{
	struct saa_video_clipinfo svci;
	uint32_t count, i;

	count = xi->xi_clist_count > SAA_MAX_CLIPS ? 
					SAA_MAX_CLIPS : xi->xi_clist_count;

	/* Check if new clist is same as already set list */
	if (xi->xi_set_clist_count == count) {
		/* XXX memcmp cannot be used directly due to ci_o* members 
		 * we don't want include them in the comparison
		 */
		for (i = 0; i < count; ++i) {
			if (memcmp(&xi->xi_set_clist[i].ci_a,
				   &xi->xi_clist[i].ci_a,
				   sizeof (struct CI_A)) != 0)
				break;
		}
		if (i >= count) {
			DMSG("cliplist: no change, count: %d\n\n", count);
			return;
		}
	}
	if (count == 0) {
		clipping_reset(vi, xi);	
	} else {
		DMSG("cliplist: changed, count: %d\n\n", count);
		memcpy(xi->xi_set_clist, xi->xi_clist,
				sizeof(struct clipinfo) * count);
		xi->xi_set_clist_count = count;
		if (vi->vi_flag_on) {
			svci.svci_clipmode = vi->vi_svs.svs_clipmode;
			generate_saa_cliplist(xi, &svci.svci_cliplist);
			if (ioctl(vi->vi_dfd, SETCLIPINFO, &svci) != 0)
				warn("Error setting SAA clipping information");
		}
	}

	if (xi->xi_dgainfo->dga_wproc_id == 0)
		xi->xi_dgainfo->dga_wproc_id =
			XtAppAddWorkProc(xi->xi_app, refresh_desktop, xi);
}


Boolean
refresh_desktop(XtPointer data)
{
	struct xinfo *xi;

	/* XXX Fix refresh code, avoid call system(...) */
#if 0
	XSetWindowAttributes attr;
	Display *d;
	Window w;
        unsigned long valuemask;

	xi = (struct xinfo *)data;
	d = XtDisplay(xi->xi_win);

	DMSG("Refreshing Desktop\n");
	valuemask = (CWBackPixel | CWBackingStore | CWSaveUnder);
	attr.background_pixel = BlackPixel(XtDisplay(xi->xi_win), 0);
	attr.backing_store = NotUseful;
	attr.save_under = False;
	w = XCreateWindow(d,
				DefaultRootWindow(d), 
				0, 0,
				xi->xi_scr_width,
				xi->xi_scr_height,
				0,
				CopyFromParent, CopyFromParent,
				CopyFromParent, valuemask,
				&attr);
	XMapWindow(d, w);
	XDestroyWindow(d, w);
	XFlush(d);

	if (xi->xi_dgainfo->dga_wproc_id != 0) {
		XtRemoveWorkProc(xi->xi_dgainfo->dga_wproc_id);
		xi->xi_dgainfo->dga_wproc_id = 0;
	}
#else
	xi = (struct xinfo *)data;
	DMSG("Refreshing Desktop\n");
	system("xrefresh");
	if (xi->xi_dgainfo->dga_wproc_id != 0) {
		XtRemoveWorkProc(xi->xi_dgainfo->dga_wproc_id);
		xi->xi_dgainfo->dga_wproc_id = 0;
	}
#endif
	return TRUE;
}


void
clipping_reset(struct vinfo *vi, struct xinfo *xi)
{
	struct saa_video_clipinfo svci;

	if (xi->xi_set_clist_count) {
		DMSG("cliplist reset\n\n");
		xi->xi_set_clist_count = 0;
		if (vi->vi_flag_on) {
			svci.svci_clipmode = vi->vi_svs.svs_clipmode;
			generate_saa_cliplist(xi, &svci.svci_cliplist);
			if (ioctl(vi->vi_dfd, SETCLIPINFO, &svci) != 0)
				warn("Error setting SAA clipping information");
		}
	}
}


void
clipping_reconfigure(struct vinfo *vi, struct xinfo *xi, Window owin, int *rval)
{
	XPoint points[4];
	XWindowAttributes wattr;
	XErrorHandler eh;
	struct clipinfo cinfo;
	Window root, win, uroot, uparent, *children;
	unsigned int count, i;
	int x, y, w, h;
	Display *d;
	Region region;


	d = XtDisplay(xi->xi_win);
	root = DefaultRootWindow(d);
	win = XtWindow(xi->xi_win);
	
	if (xi->xi_wm_parent == 0) {
		xi->xi_wm_parent = win;
		for (;;) {
			XQueryTree(d, xi->xi_wm_parent, 
					&uroot, &uparent, &children, &count);
			XFree(children);
			if (uparent == root)
				break;
			xi->xi_wm_parent = uparent;
		}
	}

	if (owin == 0) {
		XQueryTree(d, root, &uroot, &uparent, &children, &count);
		for (i = 0; i < count; ++i)
			if (children[i] == xi->xi_wm_parent)
				break;
	}

	w = xi->xi_win_width;
	h = xi->xi_win_height;
	x = xi->xi_win_xpos;
	y = xi->xi_win_ypos;

	if (x < 0) {
		w += x;
		x = 0;
	}
	if (y < 0) {
		h += y;
		y = 0;
	}
	if ((x + w) > xi->xi_scr_width)
		w = xi->xi_scr_width - x;
	if ((y + h) > xi->xi_scr_height)
		h = xi->xi_scr_height - y;	

	points[0].x = x;
	points[0].y = y;
	points[1].x = x;
	points[1].y = y + h;
	points[2].x = x + w;
	points[2].y = y + h;
	points[3].x = x + w;
	points[3].y = y;
	region = XPolygonRegion(points, 4, WindingRule);

	if (owin == 0)
		cliplist_clear(xi);

	/* Handle negative positions */
	if (xi->xi_win_xpos < 0) {
		/* No need to set ci_o* members */
		cinfo.ci_a.win = win;
		cinfo.ci_a.x = 0;
		cinfo.ci_a.y = 0;
		cinfo.ci_a.w = -xi->xi_win_xpos;
		cinfo.ci_a.h = xi->xi_win_height;
		cinfo.ci_skip = -1;
		if (owin == 0)
			cliplist_add(xi, &cinfo);
	}
	if (xi->xi_win_ypos < 0) {
		/* No need to set ci_o* members */
		cinfo.ci_a.win = win;
		cinfo.ci_a.x = 0;
		cinfo.ci_a.y = 0;
		cinfo.ci_a.w = xi->xi_win_width;
		cinfo.ci_a.h = -xi->xi_win_ypos;
		cinfo.ci_skip = -1;
		if (owin == 0)
			cliplist_add(xi, &cinfo);
	}

	if (owin != 0) {
		wattr.width = 0;
		eh = XSetErrorHandler(xgetwinattr_errorhandler);
		XGetWindowAttributes(d, owin, &wattr);
		XSetErrorHandler(eh);
		if (wattr.width != 0 && (wattr.map_state & IsViewable)) {
			*rval = XRectInRegion(region, wattr.x, wattr.y, 
						wattr.width, wattr.height);
		} else {
			*rval = RectangleOut;
		}
		XDestroyRegion(region);
		return;
	}

	DMSG("Region Rect: x: %d, y: %d, w: %d, h: %d\n", x, y, w, h);

	for (++i; i < count; ++i) {
		wattr.width = 0;
		eh = XSetErrorHandler(xgetwinattr_errorhandler);
		XGetWindowAttributes(d, children[i], &wattr);
		XSetErrorHandler(eh);
		if (wattr.width == 0 || !(wattr.map_state & IsViewable))
			continue;

		/* XXX Account for window border */
		wattr.width += wattr.border_width;
		wattr.height += wattr.border_width;

		switch (XRectInRegion(region, wattr.x, wattr.y, 
					wattr.width, wattr.height)) {
		case RectangleIn:
			cinfo.ci_a.win = children[i];
			cinfo.ci_ox = wattr.x;
			cinfo.ci_oy = wattr.y;
			cinfo.ci_ow = wattr.width;
			cinfo.ci_oh = wattr.height;
			cinfo.ci_a.x = wattr.x - xi->xi_win_xpos;
			cinfo.ci_a.y = wattr.y - xi->xi_win_ypos;
			cinfo.ci_a.w = wattr.width;
			cinfo.ci_a.h = wattr.height;
			cinfo.ci_skip = -1;
			cliplist_add(xi, &cinfo);
			break;
		case RectanglePart:
			w = wattr.width;
			h = wattr.height;
			x = wattr.x < xi->xi_win_xpos ? 
						xi->xi_win_xpos : wattr.x;
			y = wattr.y < xi->xi_win_ypos ? 
						xi->xi_win_ypos : wattr.y;
			w -= x - wattr.x;
			h -= y - wattr.y;

			if ((x + w) > (xi->xi_win_xpos + xi->xi_win_width))
				w = (xi->xi_win_xpos + xi->xi_win_width) - x;

			if ((y + h) > (xi->xi_win_ypos + xi->xi_win_height))
				h = (xi->xi_win_ypos + xi->xi_win_height) - y;

			cinfo.ci_a.win = children[i];
			cinfo.ci_ox = wattr.x;
			cinfo.ci_oy = wattr.y;
			cinfo.ci_ow = wattr.width;
			cinfo.ci_oh = wattr.height;
			cinfo.ci_a.x = x - xi->xi_win_xpos;
			cinfo.ci_a.y = y - xi->xi_win_ypos;
			cinfo.ci_a.w = w;
			cinfo.ci_a.h = h;
			cinfo.ci_skip = -1;
			cliplist_add(xi, &cinfo);
			break;
		}
	}
	XDestroyRegion(region);
	XFree(children);

	cliplist_merge(xi);
	cliplist_sort(xi);
	clipping_set_cliplist(vi, xi);
}


void
generate_saa_cliplist(struct xinfo *xi, struct saa_video_cliplist *svcl)
{
	uint32_t i;

/*
	 * XXX Figure out why this combination does not
	 * work, I suspect hw deficency
	 *
	svcl->svcl_count = 2;
	svcl->svcl_clip[0].x = 0;
	svcl->svcl_clip[0].y = 0;
	svcl->svcl_clip[0].w = 287;
	svcl->svcl_clip[0].h = 576;

	svcl->svcl_clip[1].x = 0;
	svcl->svcl_clip[1].y = 0;
	svcl->svcl_clip[1].w = 720;
	svcl->svcl_clip[1].h = 102;
*/

	svcl->svcl_count = xi->xi_set_clist_count;
	for (i = 0; i < xi->xi_set_clist_count; ++i) {
		svcl->svcl_clip[i].x = xi->xi_set_clist[i].ci_a.x;
		svcl->svcl_clip[i].y = xi->xi_set_clist[i].ci_a.y;
		svcl->svcl_clip[i].w = xi->xi_set_clist[i].ci_a.w;
		svcl->svcl_clip[i].h = xi->xi_set_clist[i].ci_a.h;
	}
}
